
import os
import sys

from numpy import get_include

RSL_MISSING_WARNING = """
==============================================================================
WARNING: RSL LIBS AND HEADERS COULD NOT BE FOUND AT THE PROVIDED LOCATION.

Py-ART will be build without bindings to the NASA TRMM RSL library but some
functionality will not be available.  If this functionality is desired please
rebuild and reinstall Py-ART after verifying:

    1. The NASA TRMM RSL library is installed and accessable.  This package
       can be obtained from:
            http://trmm-fc.gsfc.nasa.gov/trmm_gv/software/rsl/.

    2. The RSL_PATH environmental variable points to where RSL is installed
       (the libs in $RSL_PATH/lib, the headers in $RSL_PATH/include).
       Currently the RSL_PATH variable is set to: %s

==============================================================================
"""


def guess_rsl_path():
    return {'darwin': '/usr/local/trmm',
            'linux2': '/usr/local/trmm',
            'linux': '/usr/local/trmm',
            'win32': 'XXX'}[sys.platform]


def check_rsl_path(rsl_lib_path, rsl_include_path):
    """ check if the rsl path is valid, return False if invalid. """
    ext = {'darwin': 'dylib',
           'linux2': 'so',
           'linux': 'so',
           'win32': 'DLL'}[sys.platform]
    lib_file = os.path.join(rsl_lib_path, 'librsl.' + ext)
    if os.path.isfile(lib_file) is False:
        return False

    inc_file = os.path.join(rsl_include_path, 'rsl.h')
    if os.path.isfile(inc_file) is False:
        return False
    return True


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    from numpy.distutils.system_info import get_info, BlasNotFoundError
    config = Configuration('io', parent_package, top_path)
    config.add_data_dir('tests')

    # determine and verify the at RSL location
    rsl_path = os.environ.get('RSL_PATH')
    if rsl_path is None:
        rsl_path = guess_rsl_path()
    rsl_lib_path = os.path.join(rsl_path, 'lib')
    rsl_include_path = os.path.join(rsl_path, 'include')

    # build the RSL interface if RSL is installed
    if check_rsl_path(rsl_lib_path, rsl_include_path):
        config.add_extension(
            '_rsl_interface',
            sources=['_rsl_interface.c'],
            libraries=['rsl'],
            library_dirs=[rsl_lib_path],
            include_dirs=[rsl_include_path] + [get_include()],
            runtime_library_dirs=[rsl_lib_path])
    else:
        import warnings
        warnings.warn(RSL_MISSING_WARNING % (rsl_path))

    config.add_extension('_sigmetfile',
                         sources=['_sigmetfile.c'],
                         include_dirs=[get_include()])

    config.add_extension('nexrad_interpolate',
                         sources=['nexrad_interpolate.c'],
                         include_dirs=[get_include()])

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



"""
pyart.map.gates_to_grid
=======================

Generate a Cartesian grid by mapping from radar gates onto the grid.

.. autosummary::
    :toctree: generated/

    map_gates_to_grid
    _detemine_cy_weighting_func
    _find_projparams
    _parse_gatefilters
    _determine_fields
    _find_offsets
    _find_grid_params
    _parse_roi_func

"""

import numpy as np
from ..core.radar import Radar
from ..core.transforms import geographic_to_cartesian
from ..filters import GateFilter, moment_based_gate_filter

from ._gate_to_grid_map import GateToGridMapper
from ._gate_to_grid_map import RoIFunction, ConstantRoI, DistBeamRoI, DistRoI


def map_gates_to_grid(
        radars, grid_shape, grid_limits, grid_origin=None,
        grid_origin_alt=None, grid_projection=None,
        fields=None, gatefilters=False, map_roi=True,
        weighting_function='Barnes', toa=17000.0, roi_func='dist_beam',
        constant_roi=500., z_factor=0.05, xy_factor=0.02, min_radius=500.0,
        h_factor=1.0, nb=1.5, bsp=1.0, **kwargs):
    """
    Map gates from one or more radars to a Cartesian grid.

    Generate a Cartesian grid of points for the requested fields from the
    collected points from one or more radars. For each radar gate that is not
    filtered a radius of influence is calculated.  The weighted field values
    for that gate are added to all grid points within that radius.  This
    routine scaled linearly with the number of radar gates and the effective
    grid size.

    Parameters not defined below are identical to those in
    :py:func:`map_to_grid`.

    Parameters
    ----------
    roi_func : str or RoIFunction
        Radius of influence function.  A functions which takes an
        z, y, x grid location, in meters, and returns a radius (in meters)
        within which all collected points will be included in the weighting
        for that grid points. Examples can be found in the
        Typically following strings can use to specify a built in
        radius of influence function:

            * constant: constant radius of influence.
            * dist: radius grows with the distance from each radar.
            * dist_beam: radius grows with the distance from each radar
              and parameter are based of virtual beam sizes.

        A custom RoIFunction can be defined using the RoIFunction class
        and defining a get_roi method which returns the radius.  For efficient
        mapping this class should be implemented in Cython.

    Returns
    -------
    grids : dict
        Dictionary of mapped fields.  The keysof the dictionary are given by
        parameter fields.  Each elements is a `grid_size` float64 array
        containing the interpolated grid for that field.

    See Also
    --------
    grid_from_radars : Map to a grid and return a Grid object
    map_to_grid : Create grid by finding the radius of influence around each
                  grid point.

    """
    # make a tuple if passed a radar object as the first argument
    if isinstance(radars, Radar):
        radars = (radars, )

    skip_transform = False
    if len(radars) == 1 and grid_origin_alt is None and grid_origin is None:
        skip_transform = True

    if grid_origin_alt is None:
        grid_origin_alt = float(radars[0].altitude['data'])

    gatefilters = _parse_gatefilters(gatefilters, radars)
    cy_weighting_function = _detemine_cy_weighting_func(weighting_function)
    projparams = _find_projparams(grid_origin, radars, grid_projection)
    fields = _determine_fields(fields, radars)
    grid_starts, grid_steps = _find_grid_params(grid_shape, grid_limits)
    offsets = _find_offsets(radars, projparams, grid_origin_alt)
    roi_func = _parse_roi_func(roi_func, constant_roi, z_factor, xy_factor,
                               min_radius, h_factor, nb, bsp, offsets)

    # prepare grid storage arrays
    nfields = len(fields)
    grid_sum = np.zeros(grid_shape + (nfields, ), dtype=np.float32)
    grid_wsum = np.zeros(grid_shape + (nfields, ), dtype=np.float32)
    gatemapper = GateToGridMapper(
        grid_shape, grid_starts, grid_steps, grid_sum, grid_wsum)

    # project gates from each radar onto the grid
    for radar, gatefilter in zip(radars, gatefilters):

        # Copy the field data and masks.
        # TODO method that does not copy field data into new array
        shape = (radar.nrays, radar.ngates, nfields)
        field_data = np.empty(shape, dtype='float32')
        field_mask = np.empty(shape, dtype='uint8')
        for i, field in enumerate(fields):
            fdata = radar.fields[field]['data']
            field_data[:, :, i] = np.ma.getdata(fdata)
            field_mask[:, :, i] = np.ma.getmaskarray(fdata)

        # find excluded gates from the gatefilter
        if gatefilter is False:
            gatefilter = GateFilter(radar)  # include all gates
        elif gatefilter is None:
            gatefilter = moment_based_gate_filter(radar, **kwargs)
        excluded_gates = gatefilter.gate_excluded.astype('uint8')

        # calculate gate locations relative to the grid origin
        if skip_transform:
            # single radar, grid centered at radar location
            gate_x = radar.gate_x['data']
            gate_y = radar.gate_y['data']
        else:
            gate_x, gate_y = geographic_to_cartesian(
                radar.gate_longitude['data'], radar.gate_latitude['data'],
                projparams)
        gate_z = radar.gate_altitude['data'] - grid_origin_alt

        # map the gates onto the grid
        gatemapper.map_gates_to_grid(
            radar.ngates, radar.nrays, gate_z.astype('float32'),
            gate_y.astype('float32'), gate_x.astype('float32'),
            field_data, field_mask, excluded_gates,
            toa, roi_func, cy_weighting_function)

    # create and return the grid dictionary
    mweight = np.ma.masked_equal(grid_wsum, 0)
    msum = np.ma.masked_array(grid_sum, mweight.mask)
    grids = dict(
        [(f, msum[..., i] / mweight[..., i]) for i, f in enumerate(fields)])
    if map_roi:
        roi_array = np.empty(grid_shape, dtype=np.float32)
        gatemapper.find_roi_for_grid(roi_array, roi_func)
        grids['ROI'] = roi_array
    return grids


def _detemine_cy_weighting_func(weighting_function):
    """ Determine cython weight function value. """
    if weighting_function.upper() == 'CRESSMAN':
        cy_weighting_function = 1
    elif weighting_function.upper() == 'BARNES':
        cy_weighting_function = 0
    else:
        raise ValueError('unknown weighting_function')
    return cy_weighting_function


def _find_projparams(grid_origin, radars, grid_projection):
    """ Determine the projection parameter. """

    # parse grid_origin
    if grid_origin is None:
        lat = float(radars[0].latitude['data'])
        lon = float(radars[0].longitude['data'])
        grid_origin = (lat, lon)

    grid_origin_lat, grid_origin_lon = grid_origin

    # parse grid_projection
    if grid_projection is None:
            grid_projection = {
                'proj': 'pyart_aeqd', '_include_lon_0_lat_0': True}
    projparams = grid_projection.copy()
    if projparams.pop('_include_lon_0_lat_0', False):
        projparams['lon_0'] = grid_origin_lon
        projparams['lat_0'] = grid_origin_lat
    return projparams


def _parse_gatefilters(gatefilters, radars):
    """ Parse the gatefilters parameter. """
    if isinstance(gatefilters, GateFilter):
        gatefilters = (gatefilters, )  # make tuple if single filter passed
    if gatefilters is False:
        gatefilters = (False, ) * len(radars)
    if gatefilters is None:
        gatefilters = (None, ) * len(radars)
    if len(gatefilters) != len(radars):
        raise ValueError('Length of gatefilters must match length of radars')
    return gatefilters


def _determine_fields(fields, radars):
    """ Determine which field should be mapped to the grid. """
    if fields is None:
        fields = set(radars[0].fields.keys())
        for radar in radars[1:]:
            fields = fields.intersection(radar.fields.keys())
        fields = list(fields)
    return fields


def _find_offsets(radars, projparams, grid_origin_alt):
    """ Find offset between radars and grid origin. """
    # loop over the radars finding offsets from the origin
    offsets = []    # offsets from the grid origin, in meters, for each radar
    for radar in radars:
        x_disp, y_disp = geographic_to_cartesian(
            radar.longitude['data'], radar.latitude['data'], projparams)
        z_disp = float(radar.altitude['data']) - grid_origin_alt
        offsets.append((z_disp, float(y_disp), float(x_disp)))
    return offsets


def _find_grid_params(grid_shape, grid_limits):
    """ Find the starting points and step size of the grid. """
    nz, ny, nx = grid_shape
    zr, yr, xr = grid_limits
    z_start, z_stop = zr
    y_start, y_stop = yr
    x_start, x_stop = xr

    if nz == 1:
        z_step = 0.
    else:
        z_step = (z_stop - z_start) / (nz - 1.)
    if ny == 1:
        y_step = 0.
    else:
        y_step = (y_stop - y_start) / (ny - 1.)
    if nx == 1:
        x_step = 0.
    else:
        x_step = (x_stop - x_start) / (nx - 1.)

    grid_starts = (z_start, y_start, x_start)
    grid_steps = (z_step, y_step, x_step)
    return grid_starts, grid_steps


def _parse_roi_func(roi_func, constant_roi, z_factor, xy_factor, min_radius,
                    h_factor, nb, bsp, offsets):
    """ Return the Radius of influence object. """
    if not isinstance(roi_func, RoIFunction):
        if roi_func == 'constant':
            roi_func = ConstantRoI(constant_roi)
        elif roi_func == 'dist':
            roi_func = DistRoI(z_factor, xy_factor, min_radius, offsets)
        elif roi_func == 'dist_beam':
            roi_func = DistBeamRoI(h_factor, nb, bsp, min_radius, offsets)
        else:
            raise ValueError('unknown roi_func: %s' % roi_func)
    return roi_func



"""
pyart.io.nexrad_interpolate
===========================

Interpolation of NEXRAD moments from 1000 meter to 250 meter gate spacing.

.. autosummary::
    :toctree: generated/

    _fast_interpolate_scan

"""

def _fast_interpolate_scan(
        float[:, :] data, float[:] scratch_ray, float fill_value, 
        int start, int end, int moment_ngates, int linear_interp):
    """ Interpolate a single NEXRAD moment scan from 1000 m to 250 m. """
    # This interpolation scheme is only valid for NEXRAD data where a 4:1
    # (1000 m : 250 m) interpolation is needed.
    #
    # The scheme here performs a linear interpolation between pairs of gates
    # in a ray when the both of the gates are not masked (below threshold).
    # When one of the gates is masked the interpolation changes to a nearest
    # neighbor interpolation. Nearest neighbor is also performed at the end
    # points until the new range bin would be centered beyond half of the range
    # spacing of the original range.
    #
    # Nearest neighbor interpolation is performed when linear_interp is False,
    # this is equivalent to repeating each gate four times in each ray.
    #
    # No transformation of the raw data is performed prior to interpolation, so
    # reflectivity will be interpolated in dB units, velocity in m/s, etc,
    # this may not be the best method for interpolation.
    #
    # This method was adapted from Radx
    cdef int ray_num, i, interp_ngates
    cdef float gate_val, next_val, delta

    interp_ngates = 4 * moment_ngates  # number of gates interpolated

    for ray_num in range(start, end+1):

        # repeat each gate value 4 times
        for i in range(moment_ngates):
            gate_val = data[ray_num, i]
            scratch_ray[i*4 + 0] = gate_val
            scratch_ray[i*4 + 1] = gate_val
            scratch_ray[i*4 + 2] = gate_val
            scratch_ray[i*4 + 3] = gate_val

        if linear_interp:
            # linear interpolate
            for i in range(2, interp_ngates - 4, 4):
                gate_val = scratch_ray[i]
                next_val = scratch_ray[i+4]
                if gate_val == fill_value or next_val == fill_value:
                    continue
                delta = (next_val - gate_val) / 4.
                scratch_ray[i+0] = gate_val + delta * 0.5
                scratch_ray[i+1] = gate_val + delta * 1.5
                scratch_ray[i+2] = gate_val + delta * 2.5
                scratch_ray[i+3] = gate_val + delta * 3.5

        for i in range(interp_ngates):
            data[ray_num, i] = scratch_ray[i]



""" Unit Tests for Py-ART's config.py module. """

import pyart
import os
import warnings
from imp import reload

dirname = os.path.dirname(__file__)
CUSTOM_CONFIG_FILE = os.path.join(dirname, 'custom_config.py')


def test_config_functions():
    pyart.load_config(CUSTOM_CONFIG_FILE)

    assert pyart.config.get_field_name('reflectivity') == 'REF'
    assert pyart.config.get_fillvalue() == -5555.0

    metadata = pyart.config.get_metadata('azimuth')
    assert isinstance(metadata, dict)
    assert metadata['units'] == 'foo'

    assert pyart.config.get_metadata('foobar') == {}

    pyart.load_config()     # load default
    assert pyart.config.get_field_name('reflectivity') == 'reflectivity'


def test_filemetadata_custom():

    pyart.load_config(CUSTOM_CONFIG_FILE)

    # The MDV reflectivity field is mapped to velocity
    filemetadata = pyart.config.FileMetadata('mdv')
    assert filemetadata.get_field_name('DBZ_F') == 'velocity'

    # there is no foobar field
    assert filemetadata.get_field_name('foobar') is None

    # and has a custom time metadata
    time = filemetadata('time')
    assert isinstance(time, dict)
    assert 'foo' in time
    assert time['foo'] == 'bar'

    # and standard elevation metadata
    elev = filemetadata('elevation')
    assert isinstance(elev, dict)
    assert 'units' in elev
    assert elev['units'] == 'degrees'

    assert filemetadata('foobar') == {}

    # invalid filetype does no field name mapping
    filemetadata = pyart.config.FileMetadata('foo')
    assert filemetadata.get_field_name('foobar') == 'foobar'

    # additional metadata
    filemetadata = pyart.config.FileMetadata(
        'sigmet', additional_metadata={'baz': {'units': 'baz_unit'}})
    assert filemetadata('baz')['units'] == 'baz_unit'

    # file_file_names
    filemetadata = pyart.config.FileMetadata('sigmet', file_field_names=True)
    assert filemetadata.get_field_name('DBT') == 'DBT'

    # exclude fields
    filemetadata = pyart.config.FileMetadata(
        'sigmet', exclude_fields=['spectrum_width'])
    assert filemetadata.get_field_name('WIDTH2') is None


def test_init_load():

    # load the custom config and verify
    os.environ['PYART_CONFIG'] = CUSTOM_CONFIG_FILE
    reload(pyart.config)
    assert pyart.config.get_field_name('reflectivity') == 'REF'

    # starting up with an invalid config file raises a warning and
    # loads the default config
    os.environ['PYART_CONFIG'] = 'nullnullnull'
    with warnings.catch_warnings(record=True) as w:
        reload(pyart.config)
        assert len(w) > 0
    assert pyart.config.get_field_name('reflectivity') == 'reflectivity'

    # load the custom config and verify
    os.environ['PYART_CONFIG'] = CUSTOM_CONFIG_FILE
    reload(pyart.config)
    assert pyart.config.get_field_name('reflectivity') == 'REF'

    # no PYART_CONFIG loads the default
    os.environ.pop('PYART_CONFIG')
    reload(pyart.config)
    assert pyart.config.get_field_name('reflectivity') == 'reflectivity'


def test_intergration():

    pyart.load_config()     # load default
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    assert 'reflectivity' in radar.fields
    assert 'velocity' not in radar.fields

    # MDV custom maps reflectivity to velocity, and has custom time metadata
    pyart.load_config(CUSTOM_CONFIG_FILE)
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    assert 'reflectivity' not in radar.fields
    assert 'velocity' in radar.fields
    assert radar.time['foo'] == 'bar'



""" Unit Tests for Py-ART's graph/radarmapdisplay.py module. """
# execute this script to create figure_plot_radar_display_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import pyart
import numpy as np
from numpy.testing import assert_raises
from numpy.testing.decorators import skipif


# Top level Figure generating tests
@skipif(not pyart.graph.radarmapdisplay._BASEMAP_AVAILABLE)
def test_radarmapdisplay_ppi(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi_map(
        'reflectivity_horizontal', 0, colorbar_flag=True,
        title="Fancy PPI Map", mask_tuple=('reflectivity_horizontal', -100),
        resolution='c', min_lon=-100, max_lon=-93, min_lat=33, max_lat=38,
        mask_outside=True)
    display.plot_point(-95, 35, label_text='TEXT')
    display.plot_range_rings([15, 30])
    display.plot_line_geo(np.array([-95, -95]), np.array([33, 38]))
    if outfile:
        fig.savefig(outfile)
    plt.close()


# Tests of methods, these tests do not generate figures
@skipif(not pyart.graph.radarmapdisplay._BASEMAP_AVAILABLE)
def test_radarmapdisplay_auto_range():
    # test the auto_range=True function
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi_map('reflectivity_horizontal', resolution='c')
    assert round(display.basemap.latmax, 2) == 36.85
    assert round(display.basemap.latmin, 2) == 36.13
    assert round(display.basemap.lonmax, 2) == -97.15
    assert round(display.basemap.lonmin, 2) == -98.04
    plt.close()


@skipif(not pyart.graph.radarmapdisplay._BASEMAP_AVAILABLE)
def test_error_raising():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    # no basemap
    assert_raises(ValueError, display.plot_range_ring, 10)


@skipif(not pyart.graph.radarmapdisplay._BASEMAP_AVAILABLE)
def test_radardisplay_cylindrical_proj_error():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar)
    assert_raises(
        ValueError,
        display.plot_ppi_map, 'reflectivity_horizontal', projection='cyl',
        resolution='c')


if __name__ == "__main__":
    test_radarmapdisplay_ppi('figure_radarmapdisplay_ppi.png')



""" Unit Tests for Py-ART's graph/cm.py module. """


import matplotlib

import pyart
from pyart.graph import cm_colorblind


def test_colormaps_exist():
    assert isinstance(cm_colorblind.HomeyerRainbow, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.HomeyerRainbow, matplotlib.colors.Colormap)


def test_colormaps_registered():
    cmap = matplotlib.cm.get_cmap('pyart_HomeyerRainbow')
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.cm.get_cmap('pyart_HomeyerRainbow_r')
    assert isinstance(cmap, matplotlib.colors.Colormap)



"""
pyart.io.sigmet
===============

Reading and writing of Sigmet (raw format) files

.. autosummary::
    :toctree: generated/

    read_sigmet
    ymds_time_to_datetime
    _is_time_ordered_by_reversal
    _is_time_ordered_by_roll
    _is_time_ordered_by_reverse_roll
    _time_order_data_and_metadata_roll
    _time_order_data_and_metadata_reverse
    _time_order_data_and_metadata_full

"""

from __future__ import division
import datetime
import warnings

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str, _test_arguments, prepare_for_read
from ._sigmetfile import SigmetFile, bin4_to_angle, bin2_to_angle
from . import _sigmet_noaa_hh
from ..util import mean_of_two_angles_deg

SPEED_OF_LIGHT = 299793000.0


def read_sigmet(filename, field_names=None, additional_metadata=None,
                file_field_names=False, exclude_fields=None,
                time_ordered='none', full_xhdr=None, noaa_hh_hdr=None,
                debug=False, ignore_xhdr=False, ignore_sweep_start_ms=None,
                **kwargs):
    """
    Read a Sigmet (IRIS) product file.

    Parameters
    ----------
    filename : str
        Name of Sigmet (IRIS) product file to read or file-like object
        pointing to the beginning of such a file.
    field_names : dict, optional
        Dictionary mapping Sigmet data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        metadata configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the metadata configuration file will be used.
    file_field_names : bool, optional
        True to use the Sigmet data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    time_ordered : 'none', 'sequential', 'full', ...,  optional
        Parameter controlling if and how the rays are re-ordered by time.
        The default, 'none' keeps the rays ordered in the same manner as
        they appears in the Sigmet file.  'sequential' will determind and
        apply an operation which maintains a sequential ray order in elevation
        or azimuth yet orders the rays according to time.  If no operation can
        be found to accomplish this a warning is issue and the rays are
        returned in their original order.  'roll', 'reverse', and
        'reverse_and_roll' will apply that operation to the rays in order to
        place them in time order, direct use of these is not recommended.
        'full' will order the rays in strictly time increasing order,
        but the rays will likely become non-sequential, thisoption is not
        recommended unless strict time increasing order is required.
    full_xhdr : bool or None
        Flag to read in all extended headers for possible decoding. None will
        determine if extended headers should be read in automatically by
        examining the extended header type.
    noaa_hh_hdr : bool or None
        Flag indicating if the extended header should be decoded as those
        used by the NOAA Hurricane Hunters aircraft radars.  None will
        determine if the extended header is of this type automatically by
        examining the header. The `full_xhdr` parameter is set to True
        when this parameter is True.
    ignore_xhdr : bool, optional
        True to ignore all data in the extended headers if they exist.
        False, the default, extracts milliseconds precision times and other
        parameter from the extended headers if they exists in the file.
    ignore_sweep_start_ms : bool or None, optional
        True to ignore the millisecond parameter in the start time for each
        sweep, False will uses this parameter when determining the timing of
        each ray. None, the default, will ignore the millisecond sweep start
        timing only when the file does not contain extended headers or when
        the extended header has been explicity ignored using the `ignore_xhdr`
        parameter.  The TRMM RSL library ignores these times so setting this
        parameter to True is required to match the times determined when
        reading Sigmet files with :py:func:`pyart.io.read_rsl`.
        When there are not extended headers ignoring the millisecond sweep
        times provides time data which is always prior to the actual
        collection time with an error from 0 to 2 seconds.
    debug : bool, optional
        Print debug information during read.

    Returns
    -------
    radar : Radar
        Radar object

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('sigmet', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    # open the file
    sigmetfile = SigmetFile(prepare_for_read(filename), debug=debug)
    ingest_config = sigmetfile.ingest_header['ingest_configuration']
    task_config = sigmetfile.ingest_header['task_configuration']

    # determine if full extended headers should be read
    if noaa_hh_hdr is True:
        full_xhdr = True
    if full_xhdr is None:
        type_mask = task_config['task_dsp_info']['current_data_type_mask']
        if type_mask['extended_header_type'] == 2:
            full_xhdr = True
        else:
            full_xhdr = False

    # read data
    sigmet_data, sigmet_metadata = sigmetfile.read_data(full_xhdr=full_xhdr)
    first_data_type = sigmetfile.data_type_names[0]
    if first_data_type == 'XHDR':   # don't use XHDR as the first data type
        first_data_type = sigmetfile.data_type_names[1]
    sigmetfile.close()
    nsweeps, nrays, nbins = sigmet_data[first_data_type].shape

    if nsweeps == 0:
        raise IOError('File contains no readable sweep data.')

    # ignore extended header if user requested
    if ignore_xhdr:
        if 'XHDR' in sigmet_data:
            sigmet_data.pop('XHDR')
            sigmet_metadata.pop('XHDR')

    # parse the extended headers for time
    if full_xhdr and 'XHDR' in sigmet_data:
        # extract the ms timing data and store the full header for later
        # analysis, keep both in the sigmet_data dictionary so time ordering
        # and removal of missing rays is performed on these "fields"
        xhdr = sigmet_data.pop('XHDR')
        sigmet_data['XHDR'] = xhdr[:, :, :2].copy().view('i4')
        sigmet_data['XHDR_FULL'] = xhdr
        xhdr_metadata = {}
        for key in sigmet_metadata['XHDR'].keys():
            xhdr_metadata[key] = sigmet_metadata['XHDR'][key].copy()
        sigmet_metadata['XHDR_FULL'] = xhdr_metadata

    # remove missing rays from the data
    good_rays = (sigmet_metadata[first_data_type]['nbins'] != -1)
    rays_missing = (sigmet_metadata[first_data_type]['nbins'] == -1).sum()
    for field_name in sigmet_data.keys():
        sigmet_data[field_name] = sigmet_data[field_name][good_rays]
        field_metadata = sigmet_metadata[field_name]
        for key in field_metadata.keys():
            field_metadata[key] = field_metadata[key][good_rays]
    rays_per_sweep = good_rays.sum(axis=1)

    # time order
    if time_ordered == 'sequential':
        # Determine appropiate time ordering operation for the sweep and
        # determine if that operation will in fact order the times.
        # If it does not issue a warning and perform no time ordering
        if task_config['task_scan_info']['antenna_scan_mode'] == 2:
            # RHI scan
            if _is_time_ordered_by_reversal(
                    sigmet_data, sigmet_metadata, rays_per_sweep):
                time_ordered = 'reverse'
            else:
                warnings.warn('Rays not collected sequentially in time.')
                time_ordered = 'none'
        else:
            # PPI scan
            if _is_time_ordered_by_roll(
                    sigmet_data, sigmet_metadata, rays_per_sweep):
                time_ordered = 'roll'
            elif _is_time_ordered_by_reverse_roll(
                    sigmet_data, sigmet_metadata, rays_per_sweep):
                time_ordered = 'reverse_and_roll'
            else:
                warnings.warn('Rays not collected sequentially in time.')
                time_ordered = 'none'

    if time_ordered == 'full':
        _time_order_data_and_metadata_full(
            sigmet_data, sigmet_metadata, rays_per_sweep)
    if time_ordered == 'reverse':
        _time_order_data_and_metadata_reverse(
            sigmet_data, sigmet_metadata, rays_per_sweep)
    if time_ordered == 'roll':
        _time_order_data_and_metadata_roll(
            sigmet_data, sigmet_metadata, rays_per_sweep)
    if time_ordered == 'reverse_and_roll':
        # reverse followed by roll
        _time_order_data_and_metadata_reverse(
            sigmet_data, sigmet_metadata, rays_per_sweep)
        _time_order_data_and_metadata_roll(
            sigmet_data, sigmet_metadata, rays_per_sweep)

    # sweep_start_ray_index and sweep_end_ray_index
    ray_count = good_rays.sum(axis=1)
    total_rays = ray_count.sum()
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    ssri = np.cumsum(np.append([0], ray_count[:-1])).astype('int32')
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = np.cumsum(ray_count).astype('int32') - 1

    # time
    # The time of each ray collected in a Sigmet file is stored in the ray
    # header as an offset from the start of of the sweep. If the file contains
    # extended ray headers then the time is truncated to millisecond
    # precision. Without extended headers the time is truncated to second
    # precision. The time of the sweep start is recorded in the ingest data
    # header to millisecond percision.
    # The method below sets the volume starting time to the time of the first
    # sweep truncated to the nearest second so that values time['data'] are
    # always positive.  The timing of each ray is then calculated from the
    # offset of the time of the sweep from this volume starting time
    # and the offset from the sweep start to the ray recorded in the ray
    # header.  Additional, the user can select to trucate the sweep start
    # times to second precision using the `ignore_sweep_start_ms` parameter.
    # This may be desired to match libraries such as TRMM RSL, which do not
    # read the millisecond sweep start precision.  By default this
    # trucation is applied when the file contains no extended header.  This
    # ensures that the times listed occur BEFORE the ray was collected with an
    # error from 0 to 2 seconds from the recorded time.  Other methods of
    # combining the mixed percision times would result in times which may have
    # a negative error which is undesired.

    # extract times from the ray headers
    if 'XHDR' in sigmet_data:   # use time in extended headers
        tdata = sigmet_data.pop('XHDR')
        tdata = (tdata.flatten() / 1000.).astype('float64')
        sigmet_extended_header = True
    else:   # use times from the standard ray headers
        tdata = sigmet_metadata[first_data_type]['time'].astype('float64')
        sigmet_extended_header = False

    # determine sweep start times and possibly trucate them to sec. precision
    dts = [ymds_time_to_datetime(d['sweep_start_time'])
           for d in sigmetfile.ingest_data_headers[first_data_type]]
    if ignore_sweep_start_ms is None:
        ignore_sweep_start_ms = not sigmet_extended_header
    if ignore_sweep_start_ms:
        dts = [d.replace(microsecond=0) for d in dts]

    # truncate volume start time to second precision of first sweep
    dt_start = dts[0].replace(microsecond=0)

    # add sweep start times to all time values in each sweep.
    for i, dt in enumerate(dts):
        start = sweep_start_ray_index['data'][i]
        end = sweep_end_ray_index['data'][i]
        td = (dt - dt_start)    # time delta from volume start
        tdata[start:end + 1] += (td.microseconds + (td.seconds + td.days *
                                 24 * 3600) * 10**6) / 10**6

    # stores this data in the time dictionary
    time = filemetadata('time')
    time['data'] = tdata
    time['units'] = make_time_unit_str(dt_start)

    # _range
    _range = filemetadata('range')
    range_info = task_config['task_range_info']
    gate_0 = range_info['first_bin_range'] / 100.       # meters
    gate_nbin = range_info['last_bin_range'] / 100.     # meters
    gate_size = round((gate_nbin - gate_0) / (nbins))
    _range['data'] = gate_0 + gate_size * np.arange(nbins, dtype='float32')
    _range['meters_to_center_of_first_gate'] = np.array([gate_0],
                                                        dtype='float32')
    _range['meters_between_gates'] = np.array([gate_size], dtype='float32')

    # fields
    fields = {}
    for data_type_name, fdata in sigmet_data.items():
        if data_type_name == 'XHDR_FULL':
            continue
        field_name = filemetadata.get_field_name(data_type_name)
        if field_name is None:
            continue
        field_dic = filemetadata(field_name)
        field_dic['data'] = fdata.reshape(-1, nbins)
        field_dic['_FillValue'] = get_fillvalue()
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'sigmet'
    metadata['instrument_name'] = ingest_config['site_name'].strip()
    metadata['sigmet_task_name'] = (
        sigmetfile.product_hdr['product_configuration']['task_name'])

    if sigmet_extended_header:
        metadata['sigmet_extended_header'] = 'true'
    else:
        metadata['sigmet_extended_header'] = 'false'
    metadata['time_ordered'] = time_ordered
    metadata['rays_missing'] = rays_missing

    # scan_type
    if task_config['task_scan_info']['antenna_scan_mode'] == 2:
        scan_type = 'rhi'
    else:
        scan_type = 'ppi'

    # latitude
    latitude = filemetadata('latitude')
    lat = bin4_to_angle(ingest_config['latitude_radar'])
    if lat > 180.0:
        lat -= 360.0
    latitude['data'] = np.array([lat], dtype='float64')

    # longitude
    longitude = filemetadata('longitude')
    lon = bin4_to_angle(ingest_config['longitude_radar'])
    if lon > 180.0:
        lon -= 360.0
    longitude['data'] = np.array([lon], dtype='float64')

    # altitude
    altitude = filemetadata('altitude')
    alt = sigmetfile.product_hdr['product_end']['ground_height']
    altitude['data'] = np.array([alt], dtype='float64')

    # sweep_number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')

    # fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    fa = [d['fixed_angle'] for d in
          sigmetfile.ingest_data_headers[first_data_type]]
    fixed_angle['data'] = bin2_to_angle(np.array(fa)).astype('float32')

    # azimuth
    azimuth = filemetadata('azimuth')
    az0 = sigmet_metadata[first_data_type]['azimuth_0']
    az1 = sigmet_metadata[first_data_type]['azimuth_1']
    az_data = mean_of_two_angles_deg(az0, az1).astype('float32')
    az_data[az_data < 0] += 360.0
    azimuth['data'] = az_data

    # elevation
    elevation = filemetadata('elevation')
    el0 = sigmet_metadata[first_data_type]['elevation_0']
    el1 = sigmet_metadata[first_data_type]['elevation_1']
    elevation['data'] = mean_of_two_angles_deg(el0, el1).astype('float32')

    # instrument_parameters
    prt = filemetadata('prt')
    prt_mode = filemetadata('prt_mode')
    nyquist_velocity = filemetadata('nyquist_velocity')
    unambiguous_range = filemetadata('unambiguous_range')
    beam_width_h = filemetadata('radar_beam_width_h')
    beam_width_v = filemetadata('radar_beam_width_v')
    pulse_width = filemetadata('pulse_width')
    prt_ratio = filemetadata('prt_ratio')

    prt_value = 1. / sigmetfile.product_hdr['product_end']['prf']
    prt['data'] = prt_value * np.ones(total_rays, dtype='float32')

    ur_value = SPEED_OF_LIGHT * prt_value / 2.
    unambiguous_range['data'] = ur_value * np.ones(total_rays, dtype='float32')

    multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
    prf_multiplier = [1, 2, 3, 4][multi_prf_flag]
    if prf_multiplier != 1:
        prt_mode['data'] = np.array(nsweeps * ['dual'], dtype='S')
        ratio = (prf_multiplier + 1) / (prf_multiplier)  # N+1/N
        prt_ratio['data'] = ratio * np.ones(total_rays, dtype='float32')
    else:
        prt_mode['data'] = np.array(nsweeps * ['fixed'], dtype='S')
        prt_ratio['data'] = np.ones(total_rays, dtype='float32')

    wavelength_cm = sigmetfile.product_hdr['product_end']['wavelength']
    nv_value = wavelength_cm / (10000.0 * 4.0 * prt_value) * prf_multiplier
    nyquist_velocity['data'] = nv_value * np.ones(total_rays, dtype='float32')
    beam_width_h['data'] = np.array([bin4_to_angle(
        task_config['task_misc_info']['horizontal_beamwidth'])],
        dtype='float32')
    beam_width_v['data'] = np.array([bin4_to_angle(
        task_config['task_misc_info']['vertical_beamwidth'])],
        dtype='float32')
    pulse_width['data'] = np.array(
        [task_config['task_dsp_info']['pulse_width'] * 1e-8] *
        len(time['data']), dtype='float32')

    instrument_parameters = {'unambiguous_range': unambiguous_range,
                             'prt_mode': prt_mode,
                             'prt': prt,
                             'prt_ratio': prt_ratio,
                             'nyquist_velocity': nyquist_velocity,
                             'radar_beam_width_h': beam_width_h,
                             'radar_beam_width_v': beam_width_v,
                             'pulse_width': pulse_width}
    if prf_multiplier != 1:
        prf_flag = filemetadata('prf_flag')
        prf_flag['data'] = sigmet_metadata[first_data_type]['prf_flag']
        instrument_parameters['prf_flag'] = prf_flag

    # decode extended headers
    extended_header_params = {}
    if noaa_hh_hdr is None:
        type_mask = task_config['task_dsp_info']['current_data_type_mask']
        htype = type_mask['extended_header_type']
        if full_xhdr and htype == 2 and sigmet_data['XHDR_FULL'][0, 3] == 136:
            noaa_hh_hdr = True
        else:
            noaa_hh_hdr = False

    if noaa_hh_hdr:
        t = _sigmet_noaa_hh._decode_noaa_hh_hdr(
            sigmet_data['XHDR_FULL'], filemetadata, azimuth, elevation)
        (latitude, longitude, altitude, extended_header_params) = t
        metadata['platform_type'] = 'aircraft'
        # scan_type determined from the antenna_scan_mode parameters
        noaa_hh_scan_modes = {4: 'ppi', 7: 'rhi'}
        scan_mode = task_config['task_scan_info']['antenna_scan_mode']
        if scan_mode not in noaa_hh_scan_modes:
            warnings.warn("Unknown antenna_scan_mode, defaulting to 'rhi'.")
            scan_type = 'rhi'
        else:
            scan_type = noaa_hh_scan_modes[scan_mode]

    # sweep_mode
    sweep_mode = filemetadata('sweep_mode')
    if scan_type == 'ppi':
        sweep_mode['data'] = np.array(
            nsweeps * ['azimuth_surveillance'], dtype='S')
    else:
        sweep_mode['data'] = np.array(nsweeps * ['rhi'], dtype='S')

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        **extended_header_params)


def _is_time_ordered_by_reversal(data, metadata, rays_per_sweep):
    """
    Returns if volume can be time ordered by reversing some or all sweeps.
    True if the volume can be time ordered, False if not.
    """
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')
    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with no rays
        s = slice(start, start + nrays)     # slice which selects sweep
        start += nrays
        sweep_time_diff = np.diff(ref_time[s])
        if np.all(sweep_time_diff >= 0) or np.all(sweep_time_diff <= 0):
            continue
        else:
            return False
    return True


def _is_time_ordered_by_roll(data, metadata, rays_per_sweep):
    """
    Returns if volume can be time ordered by rolling some or all sweeps.
    True if the volume can be time ordered, False if not.
    """
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')
    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with no rays
        s = slice(start, start + nrays)     # slice which selects sweep
        first = ref_time[start]
        last = ref_time[start + nrays - 1]
        start += nrays
        sweep_time_diff = np.diff(ref_time[s])
        count = np.count_nonzero(sweep_time_diff < 0)
        # compare the first and last times for continuity
        if (first - last) < 0:
            count += 1
        if count != 0 and count != 1:
            return False
    return True


def _is_time_ordered_by_reverse_roll(data, metadata, rays_per_sweep):
    """
    Returns if volume can be time ordered by reversing and rolling some or all
    sweeps.  True if the volume can be time ordered, False if not.
    """
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')
    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with no rays
        s = slice(start, start + nrays)     # slice which selects sweep
        first = ref_time[start]
        last = ref_time[start + nrays - 1]
        start += nrays
        sweep_time_diff = np.diff(ref_time[s])
        if sweep_time_diff.min() < 0:   # optional reverse
            sweep_time_diff = np.diff(ref_time[s][::-1])
            first, last = last, first
        count = np.count_nonzero(sweep_time_diff < 0)
        # compare the first and last times for continuity
        if (first - last) < 0:
            count += 1
        if count != 0 and count != 1:
            return False
    return True


def _time_order_data_and_metadata_roll(data, metadata, rays_per_sweep):
    """
    Put Sigmet data and metadata in time increasing order using a roll
    operation.
    """
    # Sigmet data is stored by sweep in azimuth or elevation increasing order.
    # Time ordering PPI scans can typically be achieved by rolling the
    # ray collected first to the beginning of the sweep, which is performed
    # here.  Perfect time ordering is achieved if the rays within the sweep
    # were collected sequentially in a clockwise manner from 0 to 360 degrees
    # regardless of the first azimuth collected.
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')

    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with no rays

        s = slice(start, start + nrays)     # slice which selects sweep
        start += nrays
        # determine the number of place by which elements should be shifted.
        sweep_time = ref_time[s]
        sweep_time_diff = np.diff(sweep_time)
        if sweep_time_diff.min() >= 0:
            continue    # already time ordered
        shift = -(sweep_time_diff.argmin() + 1)

        # roll the data and metadata for each field
        for field in data.keys():
            data[field][s] = np.roll(data[field][s], shift, axis=0)
            field_metadata = metadata[field]
            for key in field_metadata.keys():
                field_metadata[key][s] = np.roll(field_metadata[key][s], shift)
    return


def _time_order_data_and_metadata_reverse(data, metadata, rays_per_sweep):
    """
    Put Sigmet data and metadata in time increasing order by reverse sweep in
    time reversed order.
    """
    # Sigmet data is stored by sweep in azimuth or elevation increasing order.
    # Time ordering RHI scans can typically be achieved by reversing the
    # ray order of sweep collected in time from 180 to 0 degrees.
    # Perfect time ordering is achieved if the rays within all sweeps
    # were collected sequentially from 0 to 180 degree or 180 to 0 degrees.
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')

    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with to few rays

        s = slice(start, start + nrays)     # slice which selects sweep
        start += nrays
        # determine the number of place by which elements should be shifted.
        sweep_time = ref_time[s]
        sweep_time_diff = np.diff(sweep_time)
        if sweep_time_diff.min() >= 0:
            continue    # already time ordered, no reversal needed
        # reverse the data and metadata for each field
        for field in data.keys():
            data[field][s] = data[field][s][::-1]
            field_metadata = metadata[field]
            for key in field_metadata.keys():
                field_metadata[key][s] = field_metadata[key][s][::-1]
    return


def _time_order_data_and_metadata_full(data, metadata, rays_per_sweep):
    """
    Put Sigmet data and metadata in time increasing order by sorting the
    times.
    """
    # Sigmet data is stored by sweep in azimuth or elevation increasing order.
    # When rays within the sweeps are collected non-sequentially or in a
    # complex manner, perfect time ordering can only be achived by sorting
    # the times themselves and reordering the rays according to this sort.
    # This ordering method should only be used as a last resort when perfect
    # time ordering is required in the output and other ordering operations
    # (roll, reverse, reverse-roll) will not order the rays correctly.
    if 'XHDR' in data:
        ref_time = data['XHDR'].copy()
        ref_time.shape = ref_time.shape[:-1]
    else:
        ref_time = metadata[metadata.keys()[0]]['time'].astype('int32')

    start = 0
    for nrays in rays_per_sweep:
        if nrays == 0 or nrays == 1:
            continue    # Do not attempt to order sweeps with no rays

        s = slice(start, start + nrays)     # slice which selects sweep
        start += nrays
        # determine the indices which sort the sweep time using a stable
        # sorting algorithm to prevent excessive azimuth scrambling.
        sweep_time = ref_time[s]
        sweep_time_diff = np.diff(ref_time[s])
        if sweep_time_diff.min() >= 0:
            continue    # already time ordered
        sort_idx = np.argsort(sweep_time, kind='mergesort')

        # sort the data and metadata for each field
        for field in data.keys():
            data[field][s] = data[field][s][sort_idx]
            field_metadata = metadata[field]
            for key in field_metadata.keys():
                field_metadata[key][s] = field_metadata[key][s][sort_idx]
    return


def ymds_time_to_datetime(ymds):
    """ Return a datetime object from a Sigmet ymds_time dictionary. """
    dt = datetime.datetime(ymds['year'], ymds['month'], ymds['day'])
    # lowest 10 bits of millisecond parameter specifies milliseconds
    microsec = 1e3 * (ymds['milliseconds'] & 0b1111111111)
    delta = datetime.timedelta(seconds=ymds['seconds'], microseconds=microsec)
    return dt + delta




# THIS FILE IS GENERATED FROM PYART SETUP.PY
short_version = '1.9.1'
version = '1.9.1'
full_version = '1.9.1'
git_revision = '746ad3fbce1c0e8b8429430284f3da9c37fea552'
release = True

if not release:
    version = full_version



"""
pyart.retrieve.kdp_proc
=======================

Module for retrieving specific differential phase (KDP) from radar total
differential phase (PSIDP) measurements. Total differential phase is a function
of propagation differential phase (PHIDP), backscatter differential phase
(DELTAHV), and the system phase offset.

.. autosummary::
    :toctree: generated/

    kdp_schneebeli
    kdp_vulpiani
    kdp_maesaka
    filter_psidp
    boundary_conditions_maesaka
    
    _kdp_estimation_backward_fixed
    _kdp_kalman_profile
    _kdp_vulpiani_profile
    _cost_maesaka
    _jac_maesaka
    _forward_reverse_phidp
    _parse_range_resolution

"""

import time
import numpy as np
import warnings
from functools import partial

from scipy import optimize, stats, interpolate, linalg, signal
from . import _kdp_proc
from ..config import get_field_name, get_metadata, get_fillvalue
from ..util import rolling_window


# Constants in the Kalman filter retrieval method (generally no need to
# modify them)
SCALERS = [0.1, 10**(-0.8), 10**(-0.6), 10**(-0.4), 10**(-0.2), 1,
           10**(0.2), 10**(0.4), 10**(0.6), 10**(0.8), 1, 10]

PADDING = 50  # Noise padding of the psidp signal (before and after signal)
SHIFT = 13  # Shifting of the final signal


def kdp_schneebeli(radar, gatefilter=None, fill_value=None, psidp_field=None,
                   kdp_field=None, phidp_field=None, band='C', rcov=0, pcov=0,
                   prefilter_psidp=False, filter_opt=None, parallel=True):
    """
    Estimates Kdp with the Kalman filter method by Schneebeli and al. (2014)
    for a set of psidp measurements.

    Parameters
    ----------
    radar : Radar
        Radar containing differential phase field.
    gatefilter : GateFilter, optional
        A GateFilter indicating radar gates that should be excluded when
        analysing differential phase measurements.
    fill_value : float, optional
        Value indicating missing or bad data in differential phase field, if
        not specified, the default in the Py-ART configuration file will be
        used
    psidp_field : str, optional
        Total differential phase field. If None, the default field name must be
        specified in the Py-ART configuration file.
    kdp_field : str, optional
        Specific differential phase field. If None, the default field name must
        be specified in the Py-ART configuration file.
    phidp_field : str, optional
        Propagation differential phase field. If None, the default field name
        must be specified in the Py-ART configuration file.
    band : char, optional
       Radar frequency band string. Accepted "X", "C", "S" (capital
       or not). The band is used to compute intercepts -c and slope b of the
       delta = b*Kdp+c relation
    rcov : 3x3 float array, optional
        Measurement error covariance matrix
    pcov  : 4x4 float array, optional
        Scaled state transition error covariance matrix
    prefilter_psidp : bool, optional
        If set, the psidp measurements will first be filtered with the
        filter_psidp method, which can improve the quality of the final Kdp
    filter_opt : dict, optional
        The arguments for the prefilter_psidp method, if empty, the defaults
        arguments of this method will be used
    parallel : bool, optional
        Flag to enable parallel computation (one core for every psidp profile)

    Returns
    -------
    kdp_dict : dict
        Retrieved specific differential phase data and metadata.
    kdp_std_dict : dict
        Estimated specific differential phase standard dev. data and metadata.
    phidpr_dict,: dict
        Retrieved differential phase data and metadata.

    References
    ----------
    Schneebeli, M., Grazioli, J., and Berne, A.: Improved Estimation
    of the Specific Differential Phase SHIFT Using a Compilation of
    Kalman Filter Ensembles, IEEE T. Geosci. Remote Sens., 52,
    5137-5149, doi:10.1109/TGRS.2013.2287017, 2014.

    """

    # create parallel computing instance
    if parallel:
        import multiprocessing as mp

        pool = mp.Pool(processes=mp.cpu_count(), maxtasksperchild=1)

    # parse fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # parse field names
    if psidp_field is None:
        psidp_field = get_field_name('differential_phase')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')

    # parse range resolution, length scale and low-pass filter constraint
    # weight
    dr = _parse_range_resolution(radar, check_uniform=True)

    # parse total differential phase measurements
    if prefilter_psidp:
        if filter_opt is None:
            filter_opt = {}
        # Assign psidp field to filter_psidp inputs
        filter_opt['psidp_field'] = psidp_field
        # Filter psidp
        psidp_o = filter_psidp(radar, **filter_opt)
    else:
        psidp_o = radar.fields[psidp_field]['data']

    # mask radar gates indicated by the gate filter
    if gatefilter is not None:
        psidp_o = np.ma.masked_where(gatefilter.gate_excluded, psidp_o)

    func = partial(_kdp_kalman_profile, dr=dr, band=band, rcov=rcov,
                   pcov=pcov)

    all_psidp_prof = list(psidp_o)

    if parallel:
        list_est = pool.map(func, all_psidp_prof)
    else:
        list_est = map(func, all_psidp_prof)

    kdp = np.zeros(psidp_o.shape) * np.nan
    kdp = np.ma.masked_array(kdp, fill_value=fill_value)

    kdp_stdev = np.zeros(psidp_o.shape) * np.nan
    kdp_stdev = np.ma.masked_array(kdp_stdev, fill_value=fill_value)

    phidp_rec = np.zeros(psidp_o.shape) * np.nan
    phidp_rec = np.ma.masked_array(phidp_rec, fill_value=fill_value)

    for i, l in enumerate(list_est):
        kdp[i, 0:len(l[0])] = l[0]
        kdp_stdev[i, 0:len(l[1])] = l[1]
        phidp_rec[i, 0:len(l[2])] = l[2]

    # Mask the estimated Kdp and reconstructed Phidp with the mask of original
    # psidp
    if isinstance(psidp_o, np.ma.masked_array):
        masked = psidp_o.mask
        kdp = np.ma.array(kdp, mask=masked, fill_value=fill_value)
        kdp_stdev = np.ma.array(kdp_stdev, mask=masked, fill_value=fill_value)
        phidp_rec = np.ma.array(phidp_rec, mask=masked, fill_value=fill_value)

    # create specific differential phase field dictionary and store data
    kdp_dict = get_metadata(kdp_field)
    kdp_dict['data'] = kdp
    # kdp_dict['valid_min'] = -1.0

    # create reconstructed differential phase field dictionary and store data
    phidpr_dict = get_metadata(phidp_field)
    phidpr_dict['data'] = phidp_rec
    # phidpr_dict['valid_min'] = 0.0

    # create specific phase stdev field dictionary and store data
    kdp_stdev_dict = {}
    kdp_stdev_dict['units'] = 'degrees/km'
    kdp_stdev_dict['standard_name'] = 'estimated KDP stdev'
    kdp_stdev_dict['long_name'] = 'Estimated stdev of spec. diff. phase (KDP)'
    kdp_stdev_dict['coordinates'] = 'elevation azimuth range'
    kdp_stdev_dict['data'] = kdp_stdev
    kdp_stdev_dict['valid_min'] = 0.0

    if parallel:
        pool.close()

    return kdp_dict, kdp_stdev_dict, phidpr_dict


def _kdp_estimation_backward_fixed(
        psidp_in, rcov, pcov_scale, f, f_transposed, h_plus,
        c1, c2, b1, b2, kdp_th, mpsidp):
    """
    Processing one profile of Psidp and estimating Kdp and Phidp
    with the KFE algorithm described in Schneebeli et al, 2014
    IEEE_TGRS. This routine estimates Kdp in the backward
    direction given a set of matrices that define the Kalman
    filter.

    Parameters
    ----------
    psidp_in : ndarray
        one-dimensional vector of length -nrg- containining the input psidp
        [degrees]
    rcov : 3x3 float array##
        Measurement error covariance matrix
    pcov_scale  : 4x4 float array
        Scaled state transition error covariance matrix
    f : 4x4 float array
        Forward state prediction matrix [4x4]
    f_transposed: 4x4 float array
        Transpose of F
    h_plus : 4x3 float array
        Measurement prediction matrix [4x3]
    c1, c2,b1,b2: floats
        the values of the intercept of the relation c  = b*Kdp - delta.
        This relation uses b1, c1 IF kdp is lower than a kdp_th and b2, c2
        otherwise kdp_th
    kdp_th: float
        the kdp threshold which separates the two Kdp - delta regime
        i.e. the power law relating delta to Kdp will be different if Kdp is
        larger or smaller than kdp_th
    mpsidp: float
        final observed value of psidp along the radial (usually also
        the max value), needed for inverting the psidp vector


    Returns
    -------
    kdp: ndarray
        filtered Kdp [degrees/km]. Same length as Psidp
    error_kdp: ndarray
        estimated error on Kdp values

    """

    # Define the input
    psidp = psidp_in

    # invert Psidp (backward estimation)
    psidp = mpsidp - psidp[::-1]
    nrg_new = len(psidp)

    # Initialize the state vector to 0
    s = np.zeros([4, 1])  # first state estimate

    # define measurement vector
    z = np.zeros([3, 1])

    # Define the identity matrix
    identity_i = np.eye(4)
    p = identity_i * 4.

    kdp = np.zeros([nrg_new])
    kdp_error = np.zeros([nrg_new])

    # Loop on all the gates and apply the filter

    for ii in range(0, nrg_new - 1):
        z[0] = psidp[ii]
        z[1] = psidp[ii + 1]

        s_pred = np.dot(f, s)  # state prediciton

        p_pred = np.dot(f, np.dot(p, f_transposed)) + \
            pcov_scale  # error prediction

        if (s_pred[0] > kdp_th):
            h_plus[2, 0] = b2
            z[2] = c2
        else:
            h_plus[2, 0] = b1
            z[2] = c1

        # as far as i see aludc is symmetrical, so i do not transpose it
        aludc = (np.dot(h_plus, (np.dot(p_pred, h_plus.T))) + rcov)

        # below we get the transposed of B_mat directly
        b_mat = np.dot(h_plus, p_pred)

        # Cholesky decomposition
        cho = linalg.cho_factor(aludc)
        k = linalg.cho_solve(cho, b_mat, check_finite=False,
                             overwrite_b=True).T

        # Update state and error
        s = np.dot(k, (np.dot(-h_plus, s_pred) + z)) + s_pred
        p = np.dot((identity_i - np.dot(k, h_plus)), p_pred)

        # Fill the output
        kdp[ii] = s[0]
        kdp_error[ii] = p[0, 0]

    # Shift
    n = len(kdp)
    dummy = np.copy(kdp)
    kdp[np.arange(SHIFT) + len(kdp) - SHIFT] = 0
    kdp[np.arange(n - 1 - SHIFT)] = dummy[np.arange(n - 1 - SHIFT) + SHIFT]

    # Reverse the estimates (backward direction)
    kdp = kdp[::-1]

    kdp_error = kdp_error[::-1]
    return kdp, kdp_error


def _kdp_estimation_forward_fixed(
        psidp_in, rcov, pcov_scale, f, f_transposed, h_plus,
        c1, c2, b1, b2, kdp_th):
    """
    Processing one profile of Psidp and estimating Kdp and Phidp
    with the KFE algorithm described in Schneebeli et al, 2014
    IEEE_TGRS. This routine estimates Kdp in the forward
    direction given a set of matrices that define the Kalman
    filter.

    Parameters
    ----------
    psidp_in : ndarray
        one-dimensional vector of length -nrg- containining the input psidp
        [degrees]
    rcov : 3x3 float array
        Measurement error covariance matrix
    pcov_scale  : 4x4 float array
        Scaled state transition error covariance matrix
    f : 4x4 float array
        Forward state prediction matrix [4x4]
    f_transposed: 4x4 float array
        Transpose of F
    h_plus : 4x3 float array*np.nan
        Measurement prediction matrix [4x3]
    c1, c2,b1,b2: floats
        the values of the intercept of the relation c  = b*Kdp - delta.
        This relation uses b1, c1 IF kdp is lower than a kdp_th and b2, c2
        otherwise kdp_th.

    Returns
    -------
    kdp: ndarray
        filtered Kdp [degrees/km]. Same length as Psidp
    phidp: ndarray
        estimated phidp (smooth psidp)
    error_kdp: ndarray
        estimated error on Kdp values

    """

    # Define the input
    psidp = np.ma.filled(psidp_in)
    nrg_new = len(psidp)

    # Initialize the state vector to 0
    s = np.zeros([4, 1])  # first state estimate

    # define measurement vector
    z = np.zeros([3, 1])

    # Define the identity matrix
    identity_i = np.eye(4)
    p = identity_i * 4.

    phidp = np.zeros([nrg_new])
    kdp = np.zeros([nrg_new])
    kdp_error = np.zeros([nrg_new])

    # Loop on all the gates and apply the filter
    for ii in range(0, nrg_new - 1):
        z[0] = psidp[ii]
        z[1] = psidp[ii + 1]

        s_pred = np.dot(f, s)  # state prediciton

        p_pred = np.dot(f, np.dot(p, f_transposed)) + \
            pcov_scale  # error prediction

        if (s_pred[0] > kdp_th):
            h_plus[2, 0] = b2
            z[2] = c2
        else:
            h_plus[2, 0] = b1
            z[2] = c1

        # as far as i see aludc is symmetrical, so i do not transpose it
        aludc = (np.dot(h_plus, (np.dot(p_pred, h_plus.T))) + rcov)

        # below we get the transposed of B_mat directly
        b_mat = np.dot(h_plus, p_pred)

        # Cholesky decomposition
        cho = linalg.cho_factor(aludc)
        k = linalg.cho_solve(
            cho,
            b_mat,
            check_finite=False,
            overwrite_b=True).T

        # Update state and error
        s = np.dot(k, (np.dot(-h_plus, s_pred) + z)) + s_pred
        p = np.dot((identity_i - np.dot(k, h_plus)), p_pred)

        # Fill the output
        kdp[ii] = s[0]
        kdp_error[ii] = p[0, 0]
        phidp[ii] = s[2]

    # Shift
    dummy = np.copy(kdp)
    kdp[np.arange(SHIFT) + len(kdp) - SHIFT] = 0
    kdp[np.arange(len(kdp) - SHIFT)
        ] = dummy[np.arange(len(kdp) - SHIFT) + SHIFT]

    return kdp, phidp, kdp_error


def _kdp_kalman_profile(psidp_in, dr, band='X', rcov=0, pcov=0):
    """
    Estimates Kdp with the Kalman filter method by Schneebeli and al. (2014)
    for a set of psidp measurements.

    Parameters
    ----------
    psidp_in : ndarray
        one-dimensional vector of length -nrg- containining the input psidp
        [degrees]
    dr : float
        Range resolution in meters.
    band : char, optional
       Radar frequency band string. Accepted "X", "C", "S" (capital
       or not). The band is used to compute intercepts -c and slope b of the
       delta = b*Kdp+c relation
    rcov : 3x3 float array, optional
        Measurement error covariance matrix
    pcov  : 4x4 float array, optional
        Scaled state transition error covariance matrix

    Returns
    -------
    kdp_dict : ndarray
        Retrieved specific differential phase data
    kdp_std_dict : ndarray
        Estimated specific differential phase standard dev. data
    phidpr_dict,: ndarray
        Retrieved differential phase data

    References
    ----------
    Schneebeli, M., Grazioli, J., and Berne, A.: Improved Estimation
    of the Specific Differential Phase Shift Using a Compilation of
    Kalman Filter Ensembles, IEEE T. Geosci. Remote Sens., 52,
    5137-5149, doi:10.1109/TGRS.2013.2287017, 2014.

    """

    dr = dr / 1000.  # Convert rad. res. to km

    # NOTE! Parameters are not checked to save as much time as possible

    # Replace missing values with nans
    psidp_in = np.ma.filled(psidp_in, np.nan)
    # Check if psidp has at least one finite value
    if not np.isfinite(psidp_in).any():
        return psidp_in, psidp_in, psidp_in  # Return the NaNs...

    # Set default of the error covariance matrices
    if not isinstance(pcov, np.ndarray):
        pcov = np.array([[(0.11 + 1.56 * dr)**2,
                          (0.11 + 1.85 * dr)**2,
                          0,
                          (0.01 + 1.1 * dr)**2],
                         [(0.11 + 1.85 * dr)**2,
                          (0.18 + 3.03 * dr)**2,
                          0,
                          (0.01 + 1.23 * dr)**2],
                         [0,
                          0,
                          0,
                          0],
                         [(0.01 + 1.1 * dr)**2,
                          (0.01 + 1.23 * dr)**2,
                          0,
                          (-0.04 + 1.27 * dr)**2]])

    if not isinstance(rcov, np.ndarray):
        rcov = np.array([[4.10625, -0.0498779, -0.0634192],
                         [-0.0498779, 4.02369, -0.0421455],
                         [-0.0634192, -0.0421455, 1.44300]])
    '''
    Define default parameters
    '''
    # Intercepts -c and slope b of delta=b*Kdp+c
    # According to the Kdp threshold selected

    if band == 'X':
        c1 = -0.054
        c2 = -6.155
        b1 = 2.3688
        b2 = 0.2734
        kdp_th = 2.5
    elif band == 'C':
        c1 = -0.036
        c2 = -1.03
        b1 = 0.53
        b2 = 0.15
        kdp_th = 2.5
    elif band == 'S':
        c1 = -0.024
        c2 = -0.15
        b1 = 0.19
        b2 = 0.019
        kdp_th = 1.1

    # Parameters for the final selection from the KF ensemble members
    fac1 = 1.2
    fac2 = 3.

    th1_comp = -0.15
    th2_comp = 0.15
    th1_final = -0.25

    # Kalman matrices
    # State matrix
    f = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1],
                  [2 * dr, 0, 0, 1]], dtype=float)
    f_transposed = f.T

    # Measurement prediction matrix--------------------------------
    # J. Grazioli modification 07.2015 --previous H_plus buggy--
    h_plus = np.array(
        [[-2 * dr, 1, 0, 1], [2 * dr, 1, 1, 0], [0, -1, 0, 0]], dtype=float)

    # Define the input
    psidp = psidp_in
    # Get indices of finite data
    real_data_ind = np.where(np.isfinite(psidp.ravel()))[0]
    offset = real_data_ind[0]

    if len(real_data_ind):
        mpsidp = psidp.ravel()[real_data_ind[-1]]
    else:
        mpsidp = np.nan

    psidp = psidp[offset:real_data_ind[-1] + 1]

    nrg = len(psidp)

    # Define the output
    kdp_filter_out = np.zeros([nrg, ])

    kdp_mat = np.zeros([nrg, 2 * len(SCALERS)])
    kdp_sim = np.zeros([nrg, len(SCALERS)])
    phidp_filter_out = np.zeros([nrg])

    '''
    Prepare a longer array with some extra gates on each side
    '''
    # add  values at the beginning and at the end of the profile
    psidp_long = np.zeros([nrg + PADDING * 2, ]) * np.nan

    nn = nrg + PADDING * 2

    noise = 2 * np.random.randn(PADDING)
    psidp_long[0:PADDING] = noise + psidp[0]
    psidp_long[nrg + PADDING: nrg + 2 * PADDING] = mpsidp + noise
    psidp_long[PADDING:nrg + PADDING] = psidp

    psidp = psidp_long

    # Get information of valid and non valid points in psidp the new psidp
    nonan = np.where(np.isfinite(psidp))[0]
    nan = np.where(np.isnan(psidp))[0]

    ranged = np.arange(0, nn)

    psidp_interp = psidp

    # interpolate
    if len(nan):
        interp = interpolate.interp1d(ranged[nonan], psidp[nonan], kind='zero')
        psidp_interp[nan] = interp(ranged[nan])
    else:
        psidp_interp = psidp

    # add noise
    if len(nan):
        psidp_interp[nan] = psidp_interp[nan] + 2 * np.random.randn(len(nan))

    # Define the final input and output
    psidp = psidp_interp

    '''
    Smallest scaler
    '''
    scaler = 10 ** (-2.)

    # Backward
    kdp_dummy_b2, _ = _kdp_estimation_backward_fixed(psidp, rcov,
                                                     pcov * scaler, f,
                                                     f_transposed, h_plus,
                                                     c1, c2, b1, b2,
                                                     kdp_th, mpsidp)
    kdp002 = kdp_dummy_b2[PADDING:nrg + PADDING]

    # Forward
    kdp_dummy_b2, _, _ = _kdp_estimation_forward_fixed(psidp, rcov,
                                                       pcov * scaler, f,
                                                       f_transposed, h_plus,
                                                       c1, c2, b1, b2, kdp_th)
    kdp002f = kdp_dummy_b2[PADDING:nrg + PADDING]

    '''
    Generate the ensemble of Kalman filters estimates in backward and
    Forward directions
    '''
    for i, sc in enumerate(SCALERS):  # Loop on scalers
        # Forward
        kdp_dummy_f2, _, _ = _kdp_estimation_forward_fixed(psidp,
                                                           rcov, pcov * sc,
                                                           f, f_transposed,
                                                           h_plus, c1, c2,
                                                           b1, b2, kdp_th)
        kdp_mat[:, 2 * i] = kdp_dummy_f2[PADDING:nrg + PADDING]

        # Backward
        kdp_dummy_b2, _ = _kdp_estimation_backward_fixed(psidp, rcov,
                                                         pcov * sc, f,
                                                         f_transposed, h_plus,
                                                         c1, c2, b1, b2,
                                                         kdp_th, mpsidp)
        kdp_mat[:, 2 * i + 1] = kdp_dummy_b2[PADDING:nrg + PADDING]

    '''
    Compile the final estimate
    '''
    # Get some reference mean values
    kdp_mean = np.nanmean(kdp_mat, axis=1)
    kdp_mean_shift = np.roll(kdp_mean, -1)
    diff_mean = kdp_mean - kdp_mean_shift

    kdp_std = np.nanstd(kdp_mat, axis=1)

    if len(diff_mean) < 4:
        size_filt = len(diff_mean)
    else:
        size_filt = 4

    diff_mean_smooth = np.convolve(diff_mean,
                                   np.ones((size_filt,)) / size_filt,
                                   mode='same')

    # Backward estimate if diff_mean greater than a defined threshold
    condi = np.where(diff_mean_smooth > th2_comp)[0]
    if len(condi):
        kdp_dummy = kdp_mat[:, np.arange(len(SCALERS)) * 2 + 1]
        kdp_sim[condi, :] = kdp_dummy[condi, :]

    # Forward estimate if diff_mean lower than a defined threshold
    condi = np.where(diff_mean_smooth < th1_comp)[0]
    if len(condi):
        kdp_dummy = kdp_mat[:, np.arange(len(SCALERS)) * 2]
        kdp_sim[condi, :] = kdp_dummy[condi, :]

    # Combination of the two in the middle
    condi = np.where(np.logical_and(diff_mean_smooth >= th1_comp,
                                    diff_mean_smooth <= th2_comp))[0]
    if len(condi):
        weight2 = (-0.5 / 0.15) * diff_mean_smooth + 0.5

        weight2 = np.tile(weight2, (len(SCALERS), 1)).T

        kdp_dummy = (1-weight2)*kdp_mat[:, np.arange(len(SCALERS))*2+1] \
            + weight2 * kdp_mat[:, np.arange(len(SCALERS)) * 2]
        kdp_sim[condi, :] = kdp_dummy[condi, :]

    # Now we reduced to 11 ensemble members: compile the final one
    kdp_mean_sim = np.nanmean(kdp_sim, axis=1)
    kdp_std_sim = np.nanstd(kdp_sim, axis=1)
    kdp_low_mean2 = np.nanmean(np.vstack((kdp002, kdp002f)).T, axis=1)

    # Get the range of ensemble members that compile
    # a final estimate

    # Lower bounds
    lower_bound = np.round(kdp_mean_sim * fac1) - np.round(kdp_std_sim * fac2)
    lower_bound = np.maximum(lower_bound, 0)
    lower_bound = np.minimum(lower_bound, len(SCALERS) - 1)

    # Upper bounds
    upper_bound = np.round(kdp_mean_sim * fac1) + np.round(kdp_std_sim * fac2)
    upper_bound = np.maximum(upper_bound, 0)
    upper_bound = np.minimum(upper_bound, len(SCALERS) - 1)

    # Final selection of the ensemble members
    for uu in range(0, nrg - 1):
        selection_vector = np.arange(
            upper_bound[uu] - lower_bound[uu] + 1) + lower_bound[uu]
        selection_vector = selection_vector.astype(int)
        kdp_filter_out[uu] = np.mean(kdp_sim[uu, selection_vector])

    # Final filtering of excessively negative values:
    # TO DO: It would be better to get rid of this filtering
    ind_lt_0 = np.where(kdp_filter_out < th1_final)[0]

    if len(ind_lt_0):
        kdp_filter_out[ind_lt_0] = kdp_low_mean2[ind_lt_0]

    # Compute phidp from Kdp
    phidp_filter_out = np.cumsum(kdp_filter_out) * 2. * dr

    phinan = np.where(np.isnan(psidp))[0]
    if len(phinan):
        phidp_filter_out[phinan] = np.nan
        kdp_filter_out[phinan] = np.nan
    phidp_filter_out[nrg - 1] = np.nan
    kdp_filter_out[nrg - 1] = np.nan

    # Pad with nan if offset > 0
    kdp_filter_out = np.pad(kdp_filter_out, (offset, 0), mode='constant',
                            constant_values=np.nan)
    kdp_std = np.pad(kdp_std, (offset, 0), mode='constant',
                     constant_values=np.nan)
    phidp_filter_out = np.pad(phidp_filter_out, (offset, 0), mode='constant',
                              constant_values=np.nan)

    return kdp_filter_out, kdp_std, phidp_filter_out


def kdp_vulpiani(radar, gatefilter=None, fill_value=None, psidp_field=None,
                 kdp_field=None, phidp_field=None, band='C', windsize=10,
                 n_iter=10, interp=False, prefilter_psidp=False,
                 filter_opt=None, parallel=False):
    """
  Estimates Kdp with the Vulpiani method for a 2D array of psidp measurements
  with the first dimension being the distance from radar and the second
  dimension being the angles (azimuths for PPI, elev for RHI).The input psidp
  is assumed to be pre-filtered (for ex. with the filter_psidp function)

    Parameters
    ----------
    radar : Radar
        Radar containing differential phase field.
    gatefilter : GateFilter, optional
        A GateFilter indicating radar gates that should be excluded when
        analysing differential phase measurements.
    fill_value : float, optional
        Value indicating missing or bad data in differential phase field, if
        not specified, the default in the Py-ART configuration file will be
        used
    psidp_field : str, optional
        Total differential phase field. If None, the default field name must be
        specified in the Py-ART configuration file.
    kdp_field : str, optional
        Specific differential phase field. If None, the default field name must
        be specified in the Py-ART configuration file.
    phidp_field : str, optional
        Propagation differential phase field. If None, the default field name
        must be specified in the Py-ART configuration file.
    band : char, optional
        Radar frequency band string. Accepted "X", "C", "S" (capital
        or not). It is used to set default boundaries for expected
        values of Kdp
    windsize : int, optional
        Size in # of gates of the range derivative window, should be even
    n_iter : int, optional
        Number of iterations of the method. Default is 10.
    interp : bool, optional
        If True, all the nans are interpolated.The advantage is that less data
        are lost (the iterations in fact are "eating the edges") but some
        non-linear errors may be introduced
    prefilter_psidp : bool, optional
        If set, the psidp measurements will first be filtered with the
        filter_psidp method, which can improve the quality of the final Kdp
    filter_opt : dict, optional
        The arguments for the prefilter_psidp method, if empty, the defaults
        arguments of this method will be used
    parallel : bool, optional
        Flag to enable parallel computation (one core for every psidp profile)

    Returns
    -------
    kdp_dict : dict
        Retrieved specific differential phase data and metadata.
    phidpr_dict,: dict
        Retrieved differential phase data and metadata.

    References
    ----------
    Gianfranco Vulpiani, Mario Montopoli, Luca Delli Passeri, Antonio G. Gioia,
    Pietro Giordano, and Frank S. Marzano, 2012: On the Use of Dual-Polarized
    C-Band Radar for Operational Rainfall Retrieval in Mountainous Areas.
    J. Appl. Meteor. Climatol., 51, 405-425, doi: 10.1175/JAMC-D-10-05024.1.

    """

    if np.mod(windsize, 2):
        warnings.warn('In the Vulpiani method, the windsize should be even ' +
                      '.Using default value, windsize = 10')
        windsize = 10

    if parallel:
        import multiprocessing as mp
        pool = mp.Pool(processes=mp.cpu_count(), maxtasksperchild=1)

    # parse fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # parse field names
    if psidp_field is None:
        psidp_field = get_field_name('differential_phase')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')

    # parse range resolution, length scale and low-pass filter constraint
    # weight
    dr = _parse_range_resolution(radar, check_uniform=True)

    # parse total differential phase measurements
    if prefilter_psidp:
        if filter_opt is None:
            filter_opt = {}
        # Assign psidp field to filter_psidp inputs
        filter_opt['psidp_field'] = psidp_field
        # Filter psidp
        psidp_o = filter_psidp(radar, **filter_opt)
    else:
        psidp_o = radar.fields[psidp_field]['data']

    # mask radar gates indicated by the gate filter
    if gatefilter is not None:
        psidp_o = np.ma.masked_where(gatefilter.gate_excluded, psidp_o)

    func = partial(_kdp_vulpiani_profile, dr=dr, windsize=windsize,
                   band=band, n_iter=n_iter, interp=interp)

    all_psidp_prof = list(psidp_o)

    if parallel:
        list_est = pool.map(func, all_psidp_prof)
    else:
        list_est = map(func, all_psidp_prof)

    kdp = np.ma.zeros(psidp_o.shape)
    kdp[:] = np.ma.masked
    kdp.set_fill_value(fill_value)

    phidp_rec = np.ma.zeros(psidp_o.shape)
    phidp_rec[:] = np.ma.masked
    phidp_rec.set_fill_value(fill_value)

    for i, l in enumerate(list_est):
        kdp[i, 0:len(l[0])] = l[0]
        phidp_rec[i, 0:len(l[1])] = l[1]

    # Mask the estimated Kdp and reconstructed Phidp with the mask of original
    # psidp
    if isinstance(psidp_o, np.ma.masked_array):
        masked = psidp_o.mask
        kdp = np.ma.array(kdp, mask=masked, fill_value=fill_value)
        phidp_rec = np.ma.array(phidp_rec, mask=masked, fill_value=fill_value)

    # create specific differential phase field dictionary and store data
    kdp_dict = get_metadata(kdp_field)
    kdp_dict['data'] = kdp
    # kdp_dict['valid_min'] = 0.0

    # create reconstructed differential phase field dictionary and store data
    phidpr_dict = get_metadata(phidp_field)
    phidpr_dict['data'] = phidp_rec
    # phidpr_dict['valid_min'] = 0.0

    if parallel:
        pool.close()

    return kdp_dict, phidpr_dict


def _kdp_vulpiani_profile(psidp_in, dr, windsize=10,
                          band='X', n_iter=10, interp=False):
    """
    Estimates Kdp with the Vulpiani method for a single profile of psidp
    measurements

    Parameters
    ----------
    psidp_in : ndarray
        Total differential phase measurements.
    dr : float
        Range resolution in meters.
    windsize : int, optional
        Size in # of gates of the range derivative window.
    band : char, optional
        Radar frequency band string. Accepted "X", "C", "S" (capital
        or not). It is used to set default boundaries for expected
        values of Kdp
    n_iter : int, optional
        Number of iterations of the method. Default is 10.
    interp : bool, optional
        If set all the nans are interpolated.The advantage is that less data
        are lost (the iterations in fact are "eating the edges") but some
        non-linear errors may be introduced

    Returns
    -------
    kdp_calc : ndarray
        Retrieved specific differential profile
    phidp_rec,: ndarray
        Retrieved differential phase profile

    """

    mask = np.ma.getmaskarray(psidp_in)
    l = windsize
    l2 = int(l/2)
    drm = dr/1000.

    if mask.all() is True:
        # Check if all elements are masked
        return psidp_in, psidp_in, psidp_in  # Return the NaNs...

    # Thresholds in kdp calculation
    if band == 'X':
        th1 = -2.
        th2 = 40.
        std_th = 5.
    elif band == 'C':
        th1 = -2.
        th2 = 20.
        std_th = 5.
    elif band == 'S':
        th1 = -2.
        th2 = 14.
        std_th = 5.
    else:
        print('Unexpected value set for the band keyword ')
        print(band)
        return None

    psidp = psidp_in
    nn = len(psidp_in)

    # Get information of valid and non valid points in psidp the new psidp
    valid = np.logical_not(mask)
    if interp:
        ranged = np.arange(0, nn)
        psidp_interp = psidp
        # interpolate
        if np.ma.is_masked(psidp):
            interp = interpolate.interp1d(ranged[valid], psidp[valid],
                                          kind='zero', bounds_error=False,
                                          fill_value=np.nan)
            psidp_interp[mask] = interp(ranged[mask])

        psidp = psidp_interp

    psidp = np.ma.filled(psidp, np.nan)
    kdp_calc = np.zeros([nn])

    # first guess
    # In the core of the profile
    kdp_calc[l2:nn-l2] = (psidp[l:nn]-psidp[0:nn-l])/(2.*l*drm)

    # set ray extremes to 0
    kdp_calc[0:l2] = 0.
    kdp_calc[nn-l2:] = 0.

    # apply thresholds
    kdp_calc[kdp_calc <= th1] = 0.
    kdp_calc[kdp_calc >= th2] = 0.

    # set all non-valid data to 0
    kdp_calc[np.isnan(kdp_calc)] = 0.

    # Remove bins with texture higher than treshold
    tex = np.ma.zeros(kdp_calc.shape)
    # compute the local standard deviation
    # (make sure that it is and odd window)
    tex_aux = np.ma.std(rolling_window(kdp_calc, l2*2+1), -1)
    tex[l2:-l2] = tex_aux
    kdp_calc[tex > std_th] = 0.

    # Loop over iterations
    for i in range(0, n_iter):
        phidp_rec = np.ma.cumsum(kdp_calc)*2.*drm

        # In the core of the profile
        kdp_calc[l2:nn-l2] = (phidp_rec[l:nn]-phidp_rec[0:nn-l])/(2.*l*drm)

        # set ray extremes to 0
        kdp_calc[0:l2] = 0.
        kdp_calc[nn-l2:] = 0.

        # apply thresholds
        kdp_calc[kdp_calc <= th1] = 0.
        kdp_calc[kdp_calc >= th2] = 0.

    # Censor Kdp where Psidp was not defined
    kdp_calc = np.ma.masked_where(mask, kdp_calc)

    # final reconstructed PhiDP from KDP
    phidp_rec = np.ma.cumsum(kdp_calc)*2.*drm

    return kdp_calc, phidp_rec


def filter_psidp(radar, psidp_field=None, rhohv_field=None, minsize_seq=5,
                 median_filter_size=7, thresh_rhohv=0.65, max_discont=90):
    """
    Filter measured psidp to remove spurious data in four steps:
         1. Censor it where Rhohv is lower than threshold
         2. Unravel angles when strong discontinuities are detected
         3. Remove very short sequences of valid data
         4. Apply a median filter on every profile

    Parameters
    ----------
    radar : Radar
        Radar containing differential phase field.

    psidp_field : str, optional
        Total differential phase field. If None, the default field name must be
        specified in the Py-ART configuration file.
    rhohv_field : str, optional
        Cross correlation ratio field. If None, the default field name must
        be specified in the Py-ART configuration file.
    minsize_seq  : integer, optional
        Minimal len (in radar gates) of sequences of valid data to be accepted
    median_filter_size : integer, optional
        Size (in radar gates) of the median filter to be applied on psidp
    thresh_rhohv : float, optional
        Censoring threshold in rhohv (gates with rhohv < thresh_rhohv)
        will be rejected
    max_discont : int, optional
        Maximum discontinuity between psidp values, default is 90 deg


    Returns
    -------
    psidp_filt : ndarray
        Filtered psidp field

    """

    # parse field names
    if psidp_field is None:
        psidp_field = get_field_name('differential_phase')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')

    # parse total differential phase and cross corr. measurements
    psidp_o = radar.fields[psidp_field]['data']
    rhohv = radar.fields[rhohv_field]['data']

    if not isinstance(psidp_o, np.ma.masked_array):
        psidp_o = np.ma.array(psidp_o, mask=np.isnan(psidp_o))

    # Initialize mask
    mask = np.ones(psidp_o.shape) * False

    # Condition on rhohv
    mask += rhohv < thresh_rhohv
    # Get original mask
    mask += psidp_o.mask

    # Remove short sequences and unwrap
    psidp_filt = np.zeros(psidp_o.shape)
    for i, psi_row in enumerate(psidp_o):
        idx = np.where(~psi_row.mask)[0]  # idx of last valid gate
        if len(idx):
            psi_row = psi_row[0:idx[-1] + 1]
            psi_row[~psi_row.mask] = np.rad2deg(
                np.unwrap(np.deg2rad(psi_row[~psi_row.mask]),
                          np.deg2rad(max_discont)))
            psi_row_with_nan = np.ma.filled(psi_row, np.nan)
            # To be sure to always have a left and right neighbour,
            # we need to pad the signal with NaN
            psi_row_with_nan = np.pad(
                psi_row_with_nan, (1, 1), 'constant', constant_values=(
                    np.nan,))
            idx = np.where(np.isfinite(psi_row_with_nan))[0]
            nan_left = idx[np.where(np.isnan(psi_row_with_nan[idx - 1]))[0]]
            nan_right = idx[np.where(np.isnan(psi_row_with_nan[idx + 1]))[0]]

            len_sub = nan_right - nan_left

            for j, l in enumerate(len_sub):
                if l < minsize_seq:
                    mask[i, nan_left[j] - 1:nan_right[j] + 1] = True

            # median filter
            psi_row = signal.medfilt(psi_row_with_nan, median_filter_size)
            psidp_filt[i, 0:len(psi_row[1:-1])] = psi_row[1:-1]

    psidp_filt = np.ma.masked_array(
        psidp_filt, mask=mask, fill_value=psidp_o.fill_value)

    return psidp_filt


# Necessary and/or potential future improvements to the KDP module:
#
# * The near and far range gate boundary conditions necessary for the Maesaka
#   et al. (2012) variational method are sensitive to the number of range gates
#   used to define both boundaries as well as outliers in these samples. This
#   may have to be further investigated and potentially improved upon.
#
# * It would be beneficial for the GateFilter class to have a method which can
#   flag radar gates above the melting layer, maybe through the user supplying
#   an estimate of the 0 deg Celsius level above mean sea level.
#
# * The backscatter differential phase parameter will likely have to be updated
#   in the future to handle various parameterizations.
#
# * The addition of an azimuthal low-pass filter (smoothness) constraint would
#   be beneficial to the variational Maesaka et al. (2012) algorithm.
#
# * Add the ability to record and/or store minimization data, e.g., current
#   iteration number, functional value and functional gradient norm value as a
#   function of iteration.


def kdp_maesaka(radar, gatefilter=None, method='cg', backscatter=None,
                Clpf=1.0, length_scale=None, first_guess=0.01,
                finite_order='low', fill_value=None, proc=1, psidp_field=None,
                kdp_field=None, phidp_field=None, debug=False, verbose=False,
                **kwargs):
    """
    Compute the specific differential phase (KDP) from corrected (e.g.,
    unfolded) total differential phase data based on the variational method
    outlined in Maesaka et al. (2012). This method assumes a monotonically
    increasing propagation differential phase (PHIDP) with increasing range
    from the radar, and therefore is limited to rainfall below the melting
    layer and/or warm clouds at weather radar frequencies (e.g., S-, C-, and
    X-band). This method currently only supports radar data with constant range
    resolution.

    Following the notation of Maesaka et al. (2012), the primary control
    variable k is proportional to KDP,

                                k**2 = 2 * KDP * dr

    which, because of the square, assumes that KDP always takes a positive
    value.

    Parameters
    ----------
    radar : Radar
        Radar containing differential phase field.
    gatefilter : GateFilter
        A GateFilter indicating radar gates that should be excluded when
        analysing differential phase measurements.
    method : str, optional
        Type of scipy.optimize method to use when minimizing the cost
        functional. The default method uses a nonlinear conjugate gradient
        algorithm. In Maesaka et al. (2012) they use the Broyden-Fletcher-
        Goldfarb-Shanno (BFGS) algorithm, however for large functional size
        (e.g., 100K+ variables) this algorithm is considerably slower than a
        conjugate gradient algorithm.
    backscatter : optional
        Define the backscatter differential phase. If None, the backscatter
        differential phase is set to zero for all range gates. Note that
        backscatter differential phase can be parameterized using attentuation
        corrected differential reflectivity.
    Clpf : float, optional
        The low-pass filter (radial smoothness) constraint weight as in
        equation (15) of Maesaka et al. (2012).
    length_scale : float, optional
        Length scale in meters used to bring the dimension and magnitude of the
        low-pass filter cost functional in line with the observation cost
        functional. If None, the length scale is set to the range resolution.
    first_guess : float, optional
        First guess for control variable k. Since k is proportional to the
        square root of KDP, the first guess should be close to zero to signify
        a KDP field close to 0 deg/km everywhere. However, the first guess
        should not be exactly zero in order to avoid convergence criteria after
        the first iteration. In fact it is recommended to use a value closer to
        one than zero.
    finite_order : 'low' or 'high', optional
        The finite difference accuracy to use when computing derivatives.
    maxiter : int, optional
        Maximum number of iterations to perform during cost functional
        minimization. The maximum number of iterations are only performed if
        convergence criteria are not met. For variational schemes such as this
        one, it is generally not recommended to try and achieve convergence
        criteria since the values of the cost functional and/or its gradient
        norm are somewhat arbitrary.
    fill_value : float, optional
        Value indicating missing or bad data in differential phase field.
    proc : int, optional
        The number of parallel threads (CPUs) to use. Currently no
        multiprocessing capability exists.
    psidp_field : str, optional
        Total differential phase field. If None, the default field name must be
        specified in the Py-ART configuration file.
    kdp_field : str, optional
        Specific differential phase field. If None, the default field name must
        be specified in the Py-ART configuration file.
    phidp_field : str, optional
        Propagation differential phase field. If None, the default field name
        must be specified in the Py-ART configuration file.
    debug : bool, optional
        True to print debugging information, False to suppress.
    verbose : bool, optional
        True to print relevant information, False to suppress.

    Returns
    -------
    kdp_dict : dict
        Retrieved specific differential phase data and metadata.
    phidpf_dict, phidpr_dict : dict
        Retrieved forward and reverse direction propagation differential phase
        data and metadata.

    References
    ----------
    Maesaka, T., Iwanami, K. and Maki, M., 2012: "Non-negative KDP Estimation
    by Monotone Increasing PHIDP Assumption below Melting Layer". The Seventh
    European Conference on Radar in Meteorology and Hydrology.

    """

    # parse fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # parse field names
    if psidp_field is None:
        psidp_field = get_field_name('differential_phase')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')

    # parse range resolution, length scale and low-pass filter constraint
    # weight
    dr = _parse_range_resolution(radar, check_uniform=True, verbose=verbose)

    # parse length scale
    if length_scale is None:
        length_scale = dr

    # parse low-pass filter constraint weight
    # this brings the dimensions and magnitude of the low-pass filter
    # constraint in line with the differential phase measurement constraint
    Clpf *= length_scale**4

    # parse near and far range gate propagation differential phase boundary
    # conditions
    bcs = boundary_conditions_maesaka(
        radar, gatefilter=gatefilter, psidp_field=psidp_field, debug=debug,
        verbose=verbose, **kwargs)
    phi_near, phi_far = bcs[:2]
    idx_near, idx_far = bcs[4:]

    # parse total differential phase measurements
    psidp_o = radar.fields[psidp_field]['data']

    if debug:
        N = np.ma.count(psidp_o)
        print('Sample size before filtering: {}'.format(N))

    # mask radar gates indicated by the gate filter
    if gatefilter is not None:
        psidp_o = np.ma.masked_where(gatefilter.gate_excluded, psidp_o)

    # mask any radar gates which are closer (further) than the near (far)
    # boundary condition ranges
    for ray in range(radar.nrays):
        psidp_o[ray, :idx_near[ray]] = np.ma.masked
        psidp_o[ray, idx_far[ray] + 1:] = np.ma.masked

    if debug:
        N = np.ma.count(psidp_o)
        print('Sample size after filtering: {}'.format(N))

    # parse differential phase measurement weight
    Cobs = np.logical_not(np.ma.getmaskarray(psidp_o)).astype(psidp_o.dtype)
    psidp_o = np.ma.filled(psidp_o, fill_value)

    # parse backscattered differential phase data
    if backscatter is None:
        dhv = np.zeros_like(psidp_o, subok=False)

    # parse solver options
    options = {
        'maxiter': kwargs.get('maxiter', 50),
        'gtol': kwargs.get('gtol', 1.0e-5),
        'disp': verbose,
    }
    if debug:
        optimize.show_options(solver='minimize', method=method)

    # parse initial conditions (first guess)
    x0 = np.zeros_like(psidp_o, subok=False).flatten()
    x0.fill(first_guess)

    if verbose:
        print('Cost functional size: {}'.format(x0.size))

    # define arguments for cost functional and its Jacobian (gradient)
    args = (psidp_o, [phi_near, phi_far],
            dhv, dr, Cobs, Clpf,
            finite_order, fill_value,
            proc, debug, verbose)

    if debug:
        start = time.time()

    # minimize the cost functional
    xopt = optimize.minimize(
        _cost_maesaka, x0, args=args, method=method, jac=_jac_maesaka,
        hess=None, hessp=None, bounds=None, constraints=None, callback=None,
        options=options)

    if debug:
        elapsed = time.time() - start
        print('Elapsed time for minimization: {:.0f} sec'.format(elapsed))

    # parse control variables from optimized result
    k = xopt.x.reshape(psidp_o.shape)

    # compute specific differential phase from control variable k in deg/km
    kdp = k**2 / (2.0 * dr) * 1000.0

    if debug:
        print('Min retrieved KDP: {:.2f} deg/km'.format(kdp.min()))
        print('Max retrieved KDP: {:.2f} deg/km'.format(kdp.max()))
        print('Mean retrieved KDP: {:.2f} deg/km'.format(kdp.mean()))

    # create specific differential phase field dictionary and store data
    kdp_dict = get_metadata(kdp_field)
    kdp_dict['data'] = kdp
    kdp_dict['valid_min'] = 0.0
    kdp_dict['Clpf'] = Clpf

    # compute forward and reverse direction propagation differential phase
    phidp_f, phidp_r = _forward_reverse_phidp(
        k, [phi_near, phi_far], verbose=verbose)

    # create forward direction propagation differential phase field dictionary
    # and store data
    phidpf_dict = get_metadata(phidp_field)
    phidpf_dict['data'] = phidp_f
    phidpf_dict['comment'] = 'Retrieved in forward direction'

    # create reverse direction propagation differential phase field dictionary
    # and store data
    phidpr_dict = get_metadata(phidp_field)
    phidpr_dict['data'] = phidp_r
    phidpr_dict['comment'] = 'Retrieved in reverse direction'

    return kdp_dict, phidpf_dict, phidpr_dict


def boundary_conditions_maesaka(
        radar, gatefilter=None, n=20, psidp_field=None, debug=False,
        verbose=False, **kwargs):
    """
    Determine near range gate and far range gate propagation differential phase
    boundary conditions. This follows the method outlined in Maesaka et al.
    (2012), except instead of using the mean we use the median which is less
    susceptible to outliers. This function can also be used to estimate the
    system phase offset.

    Parameters
    ----------
    radar : Radar
        Radar containing total differential phase measurements.
    gatefilter : GateFilter
        A GateFilter indicating radar gates that should be excluded when
        analysing differential phase measurements.
    n : int, optional
        The number of range gates necessary to define the near and far range
        gate boundary conditions. Maesaka et al. (2012) uses a value of 30. If
        this value is too small then a spurious spike in specific differential
        phase close to the radar may be retrieved.
    check_outliers : bool, optional
        True to check for near range gate boundary condition outliers. Outliers
        near the radar are primarily the result of ground clutter returns.
    psidp_field : str, optional
        Field name of total differential phase. If None, the default field name
        must be specified in the Py-ART configuration file.
    debug : bool, optional
        True to print debugging information, False to suppress.
    verbose : bool, optional
        True to print relevant information, False to suppress.

    Returns
    -------
    phi_near : ndarray
        The near range differential phase boundary condition for each ray.
    phi_far : ndarray
        The far range differential phase boundary condition for each ray.
    range_near : ndarray
        The near range gate in meters for each ray.
    range_far : ndarray
        The far range gate in meters for each ray.
    idx_near : ndarray
        Index of nearest range gate for each ray.
    idx_far : ndarray
        Index of furthest range gate for each ray.

    """

    # parse field names
    if psidp_field is None:
        psidp_field = get_field_name('differential_phase')

    # parse differential phase measurements
    psidp = radar.fields[psidp_field]['data']
    if gatefilter is not None:
        psidp = np.ma.masked_where(gatefilter.gate_excluded, psidp)

    # find contiguous unmasked data along each ray
    slices = np.ma.notmasked_contiguous(psidp, axis=1)
    if debug:
        N = sum([len(slc) for slc in slices if hasattr(slc, '__len__')])
        print('Total number of unique non-masked regions: {}'.format(N))

    phi_near = np.zeros(radar.nrays, dtype=psidp.dtype)
    phi_far = np.zeros(radar.nrays, dtype=psidp.dtype)
    range_near = np.zeros(radar.nrays, dtype=psidp.dtype)
    range_far = np.zeros(radar.nrays, dtype=psidp.dtype)
    idx_near = np.zeros(radar.nrays, dtype=np.int32)
    idx_far = np.zeros(radar.nrays, dtype=np.int32)

    for ray, regions in enumerate(slices):

        # check if all range gates are missing
        if regions is None:
            continue

        # check if all range gates available, i.e., no masked data
        if isinstance(regions, slice):
            regions = [regions]

        # near range gate boundary condition
        for slc in regions:

            # check if enough samples exist in slice
            if slc.stop - slc.start >= n:

                # parse index and range of nearest range gate
                idx = slice(slc.start, slc.start + n)
                idx_near[ray] = idx.start
                range_near[ray] = radar.range['data'][idx][0]

                # parse data for linear regression and compute slope
                x = radar.range['data'][idx]
                y = psidp[ray, idx]
                slope = stats.linregress(x, y)[0]

                # if linear regression slope is positive, set near range gate
                # boundary condition to first differential phase measurement,
                # otherwise use the median value
                if slope > 0.0:
                    phi_near[ray] = y[0]
                else:
                    phi_near[ray] = np.median(y)
                break

            else:
                continue

        # far range gate boundary condition
        for slc in reversed(regions):

            if slc.stop - slc.start >= n:

                # parse index and range of furthest range gate
                idx = slice(slc.stop - n, slc.stop)
                idx_far[ray] = idx.stop
                range_far[ray] = radar.range['data'][idx][-1]

                # parse data for linear regression and compute slope
                x = radar.range['data'][idx]
                y = psidp[ray, idx]
                slope = stats.linregress(x, y)[0]

                # if linear regression slope is positive, set far range gate
                # boundary condition to last differential phase measurement,
                # otherwise use the median value
                if slope > 0.0:
                    phi_far[ray] = y[-1]
                else:
                    phi_far[ray] = np.median(y)
                break

            else:
                continue

    # check for outliers in the near range boundary conditions, e.g., ground
    # clutter can introduce spurious values in certain rays
    if kwargs.get('check_outliers', True):

        # do not include missing values in the analysis
        phi_near_valid = phi_near[phi_near != 0.0]

        # bin and count near range boundary condition values, i.e., create
        # a distribution of values
        # the default bin width is 5 deg
        counts, edges = np.histogram(
            phi_near_valid, bins=144, range=(-360, 360), density=False)

        # assume that the maximum counts corresponds to the maximum in the
        # system phase distribution
        # this assumption breaks down if there are a significant amount of
        # near range boundary conditions characterized by ground clutter
        system_phase_peak_left = edges[counts.argmax()]
        system_phase_peak_right = edges[counts.argmax() + 1]

        if debug:
            print('Peak of system phase distribution: {:.0f} deg'.format(
                  system_phase_peak_left))

        # determine left edge location of system phase distribution
        # we consider five counts or less to be insignificant
        is_left_side = np.logical_and(
            edges[:-1] < system_phase_peak_left, counts <= 5)
        left_edge = edges[:-1][is_left_side][-1]

        # determine right edge location of system phase distribution
        # we consider five counts or less to be insignificant
        is_right_side = np.logical_and(
            edges[1:] > system_phase_peak_right, counts <= 5)
        right_edge = edges[1:][is_right_side][0]

        if debug:
            print('Left edge of system phase distribution: {:.0f} deg'.format(
                  left_edge))
            print('Right edge of system phase distribution: {:.0f} deg'.format(
                  right_edge))

        # define the system phase offset as the median value of the system
        # phase distriubion
        is_system_phase = np.logical_or(
            phi_near_valid >= left_edge, phi_near_valid <= right_edge)
        system_phase_offset = np.median(phi_near_valid[is_system_phase])

        if debug:
            print('Estimated system phase offset: {:.0f} deg'.format(
                  system_phase_offset))

        for ray, bc in enumerate(phi_near):

            # if near range boundary condition does not draw from system phase
            # distribution then set it to the system phase offset
            if bc < left_edge or bc > right_edge:
                phi_near[ray] = system_phase_offset

    # check for unphysical boundary conditions, i.e., propagation differential
    # phase should monotonically increase from the near boundary to the far
    # boundary
    is_unphysical = phi_far - phi_near < 0.0
    phi_far[is_unphysical] = phi_near[is_unphysical]
    if verbose:
        N = is_unphysical.sum()
        print('Rays with unphysical boundary conditions: {}'.format(N))

    return phi_near, phi_far, range_near, range_far, idx_near, idx_far


def _cost_maesaka(x, psidp_o, bcs, dhv, dr, Cobs, Clpf, finite_order,
                  fill_value, proc, debug=False, verbose=False):
    """
    Compute the value of the cost functional similar to equations (12)-(15) in
    Maesaka et al. (2012).

    Parameters
    ----------
    x : ndarray
        Analysis vector containing control variable k.
    psidp_o : ndarray
        Total differential phase measurements.
    bcs : array_like
        The near and far range gate propagation differential phase boundary
        conditions.
    dhv : ndarray
        Backscatter differential phase.
    dr : float
        Range resolution in meters.
    Cobs : ndarray
        The differential phase measurement constraint weights. The weight
        should vanish where no differential phase measurements are available.
    Clpf : float
        The low-pass filter (radial smoothness) constraint weight as in
        equation (15) of Maesaka et al. (2012).
    finite_order : 'low' or 'high'
        The finite difference accuracy to use when computing derivatives.
    fill_value : float
        Value indicating missing or bad data in radar field data.
    proc : int
        The number of parallel threads (CPUs) to use.
    debug : bool, optional
        True to print debugging information, False to suppress.
    verbose : bool, optional
        True to print progress information, False to suppress.

    Returns
    -------
    J : float
        Value of total cost functional.

    """

    # parse control variable k from analysis vector
    nr, ng = psidp_o.shape
    k = x.reshape(nr, ng)

    # parse near and far range gate boundary conditions
    phi_near, phi_far = bcs

    # compute forward direction propagation differential phase from control
    # variable k
    phi_fa = np.zeros_like(k, subok=False)
    phi_fa[:, 1:] = np.cumsum(k[:, :-1]**2, axis=1)

    # compute reverse direction propagation differential phase from control
    # variable k
    phi_ra = np.zeros_like(k, subok=False)
    phi_ra[:, :-1] = np.cumsum(k[:, :0:-1]**2, axis=1)[:, ::-1]

    # compute forward and reverse propagation differential phase
    # from total differential phase observations
    phi_fo = psidp_o - dhv - phi_near[:, np.newaxis].repeat(ng, axis=1)
    phi_ro = phi_far[:, np.newaxis].repeat(ng, axis=1) - psidp_o + dhv

    # cost: forward direction differential phase observations
    Jof = 0.5 * np.sum(Cobs * (phi_fa - phi_fo)**2)

    # cost: reverse direction differential phase observations
    Jor = 0.5 * np.sum(Cobs * (phi_ra - phi_ro)**2)

    # prepare control variable k for Cython function
    k = np.ascontiguousarray(k, dtype=np.float64)

    # compute low-pass filter term, i.e., second order derivative of k with
    # respect to range
    d2kdr2 = np.empty_like(k)
    _kdp_proc.lowpass_maesaka_term(k, dr, finite_order, d2kdr2)

    # cost: low-pass filter, i.e., radial smoothness
    Jlpf = 0.5 * np.sum(Clpf * (d2kdr2)**2)

    # compute value of total cost functional
    J = Jof + Jor + Jlpf

    if verbose:
        print('Forward direction observation cost : {:1.3e}'.format(Jof))
        print('Reverse direction observation cost : {:1.3e}'.format(Jor))
        print('Low-pass filter cost ............. : {:1.3e}'.format(Jlpf))
        print('Total cost ....................... : {:1.3e}'.format(J))

    return J


def _jac_maesaka(x, psidp_o, bcs, dhv, dr, Cobs, Clpf, finite_order,
                 fill_value, proc, debug=False, verbose=False):
    """
    Compute the Jacobian (gradient) of the cost functional similar to equations
    (16)-(18) in Maesaka et al. (2012).

    Parameters
    ----------
    x : ndarray
        Analysis vector containing control variable k.
    psidp_o : ndarray
        Total differential phase measurements.
    bcs : array_like
        The near and far range gate propagation differential phase boundary
        conditions.
    dhv : ndarray
        Backscatter differential phase.
    dr : float
        Range resolution in meters.
    Cobs : ndarray
        The differential phase measurement constraint weights. The weight
        should vanish where no differential phase measurements are available.
    Clpf : float
        The low-pass filter (radial smoothness) constraint weight as in
        equation (15) of Maesaka et al. (2012).
    finite_order : 'low' or 'high'
        The finite difference accuracy to use when computing derivatives.
    fill_value : float
        Value indicating missing or bad data in radar field data.
    proc : int
        The number of parallel threads (CPUs) to use.
    debug : bool, optional
        True to print debugging information, False to suppress.
    verbose : bool, optional
        True to print progress information, False to suppress.

    Returns
    -------
    jac : ndarray
        Jacobian of the cost functional.

    """

    # parse control variable k from analysis vector
    nr, ng = psidp_o.shape
    k = x.reshape(nr, ng)

    # parse near and far range gate boundary conditions
    phi_near, phi_far = bcs

    # compute forward direction propagation differential phase from control
    # variable k
    phi_fa = np.zeros_like(k, subok=False)
    phi_fa[:, 1:] = np.cumsum(k[:, :-1]**2, axis=1)

    # compute reverse direction propagation differential phase from control
    # variable k
    phi_ra = np.zeros_like(k, subok=False)
    phi_ra[:, :-1] = np.cumsum(k[:, :0:-1]**2, axis=1)[:, ::-1]

    # compute forward and reverse propagation differential phase
    # from total differential phase observations
    phi_fo = psidp_o - dhv - phi_near[:, np.newaxis].repeat(ng, axis=1)
    phi_ro = phi_far[:, np.newaxis].repeat(ng, axis=1) - psidp_o + dhv

    # prepare control variable k for Cython functions
    k = np.ascontiguousarray(k, dtype=np.float64)

    # cost: forward direction differential phase observations
    dJofdk = np.zeros_like(k, subok=False)
    dJofdk[:, :-1] = 2.0 * k[:, :-1] * np.cumsum(
        (Cobs[:, 1:] * (phi_fa[:, 1:] - phi_fo[:, 1:]))[:, ::-1],
        axis=1)[:, ::-1]

    # cost: reverse direction differential phase observations
    dJordk = np.zeros_like(k, subok=False)
    dJordk[:, 1:] = 2.0 * k[:, 1:] * np.cumsum(
        (Cobs[:, :-1] * (phi_ra[:, :-1] - phi_ro[:, :-1])), axis=1)

    # compute low-pass filter term, i.e., second order derivative of k with
    # respect to range
    d2kdr2 = np.empty_like(k)
    _kdp_proc.lowpass_maesaka_term(k, dr, finite_order, d2kdr2)

    # compute gradients of Jlpf with respect to the control variable k
    dJlpfdk = np.empty_like(d2kdr2)
    _kdp_proc.lowpass_maesaka_jac(d2kdr2, dr, Clpf, finite_order, dJlpfdk)

    # sum control variable derivative components
    dJdk = dJofdk + dJordk + dJlpfdk
    jac = dJdk.flatten()

    # compute the vector norm of the Jacobian
    if verbose:
        mag = np.linalg.norm(jac, ord=None, axis=None)
        print('Vector norm of Jacobian: {:1.3e}'.format(mag))

    return jac


def _forward_reverse_phidp(k, bcs, verbose=False):
    """
    Compute the forward and reverse direction propagation differential phases
    from the control variable k and boundary conditions following equations (1)
    and (7) in Maesaka et al. (2012).

    Parameters
    ----------
    k : ndarray
        Control variable k of the Maesaka et al. (2012) method. The control
        variable k is proportional to the square root of specific differential
        phase.
    bcs : array_like
        The near and far range gate boundary conditions.
    verbose : bool, optional
        True to print relevant information, False to suppress.

    Returns
    -------
    phidp_f : ndarray
        Forward direction propagation differential phase.
    phidp_r : ndarray
        Reverse direction propagation differential phase.

    """

    # parse near and far range gate boundary conditions
    nr, ng = k.shape
    phi_near, phi_far = bcs

    # compute forward direction propagation differential phase
    phi_f = np.zeros_like(k, subok=False)
    phi_f[:, 1:] = np.cumsum(k[:, :-1]**2, axis=1)
    phidp_f = phi_f + phi_near[:, np.newaxis].repeat(ng, axis=1)

    # compute reverse direction propagation differential phase
    phi_r = np.zeros_like(k, subok=False)
    phi_r[:, :-1] = np.cumsum(k[:, :0:-1]**2, axis=1)[:, ::-1]
    phidp_r = phi_far[:, np.newaxis].repeat(ng, axis=1) - phi_r

    # check quality of retrieval by comparing forward and reverse directions
    if verbose:
        phidp_mbe = np.ma.mean(phidp_f - phidp_r)
        phidp_mae = np.ma.mean(np.abs(phidp_f - phidp_r))
        print('Forward-reverse PHIDP MBE: {:.2f} deg'.format(phidp_mbe))
        print('Forward-reverse PHIDP MAE: {:.2f} deg'.format(phidp_mae))

    return phidp_f, phidp_r


def _parse_range_resolution(
        radar, check_uniform=True, atol=1.0, verbose=False):
    """
    Parse the radar range gate resolution.

    Parameters
    ----------
    radar : Radar
        Radar containing range data.
    check_uniform : bool, optional
        True to check if all range gates are equally spaced, and if so return
        a scalar value for range resolution. If False, the resolution between
        each range gate is returned.
    atol : float, optional
        The absolute tolerance in meters allowed for discrepancies in range
        gate spacings. Only applicable when check_uniform is True. This
        parameter may be necessary to catch instances where range gate spacings
        differ by a few meters or so.
    verbose : bool, optional
        True to print the range gate resolution. Only valid if check_uniform is
        True.

    Returns
    -------
    dr : float or ndarray
        The radar range gate spacing in meters.

    """

    # parse radar range gate spacings
    dr = np.diff(radar.range['data'], n=1)

    # check for uniform resolution
    if check_uniform and np.allclose(np.diff(dr), 0.0, atol=atol):
        dr = dr[0]
        if verbose:
            print('Range resolution: {:.2f} m'.format(dr))
    else:
        raise ValueError('Radar gate spacing is not uniform')

    return dr
    


""" Unit Tests for Py-ART's core/wind_profile.py module. """

from __future__ import print_function

import numpy as np
from numpy.testing import assert_almost_equal, assert_raises

from pyart.core import HorizontalWindProfile


def test_horizontalwindprofile_class():
    height = np.arange(5)
    speed = np.ones(5)
    direction = np.array([0, 90, 180, 270, 45])
    hprofile = HorizontalWindProfile(height, speed, direction)

    assert_almost_equal(hprofile.height, [0, 1, 2, 3, 4])
    assert_almost_equal(hprofile.speed, [1, 1, 1, 1, 1])
    assert_almost_equal(hprofile.direction, [0, 90, 180, 270, 45])
    assert_almost_equal(hprofile.u_wind, [0, -1, 0, 1, -0.707], 3)
    assert_almost_equal(hprofile.v_wind, [-1, 0, 1, 0, -0.707], 3)


def test_horizontalwindprofile_from_u_and_v():
    height = np.arange(5)
    u_wind = [0, -1, 0, 1, -np.sqrt(2)/2.]
    v_wind = [-1, 0, 1, 0, -np.sqrt(2)/2.]
    hprofile = HorizontalWindProfile.from_u_and_v(height, u_wind, v_wind)

    assert_almost_equal(hprofile.height, [0, 1, 2, 3, 4])
    assert_almost_equal(hprofile.speed, [1, 1, 1, 1, 1])
    assert_almost_equal(hprofile.direction, [0, 90, 180, 270, 45])
    assert_almost_equal(hprofile.u_wind, [0, -1, 0, 1, -0.707], 3)
    assert_almost_equal(hprofile.v_wind, [-1, 0, 1, 0, -0.707], 3)


def test_horizontalwindprofile_error():
    height = [1, 2]
    speed = [1, 1]
    direction = [1, 2, 3]
    assert_raises(ValueError, HorizontalWindProfile, height, speed, direction)


def test_horizontalwindprofile_location_error():
    height = [1, 2]
    speed = [1, 1]
    direction = [1, 2]
    lat = [1]
    assert_raises(ValueError, HorizontalWindProfile, height, speed, direction,
                  latitude=lat, longitude=None)



"""
pyart.graph.common
==================

Common graphing routines.

.. autosummary::
    :toctree: generated/

    parse_ax
    parse_ax_fig
    parse_cmap
    parse_vmin_vmax
    parse_lon_lat
    generate_colorbar_label
    generate_field_name
    generate_radar_name
    generate_grid_name
    generate_radar_time_begin
    generate_radar_time_sweep
    generate_grid_time_begin
    generate_filename
    generate_grid_filename
    generate_title
    generate_grid_title
    generate_longitudinal_level_title
    generate_latitudinal_level_title
    generate_vpt_title
    generate_ray_title
    set_limits

"""

import matplotlib.pyplot as plt
from netCDF4 import num2date

from ..config import get_field_colormap, get_field_limits


########################
# Common radar methods #
########################


def parse_ax(ax):
    """ Parse and return ax parameter. """
    if ax is None:
        ax = plt.gca()
    return ax


def parse_ax_fig(ax, fig):
    """ Parse and return ax and fig parameters. """
    if ax is None:
        ax = plt.gca()
    if fig is None:
        fig = plt.gcf()
    return ax, fig


def parse_cmap(cmap, field=None):
    """ Parse and return the cmap parameter. """
    if cmap is None:
        cmap = get_field_colormap(field)
    return cmap


def parse_vmin_vmax(container, field, vmin, vmax):
    """ Parse and return vmin and vmax parameters. """
    field_dict = container.fields[field]
    field_default_vmin, field_default_vmax = get_field_limits(field)
    if vmin is None:
        if 'valid_min' in field_dict:
            vmin = field_dict['valid_min']
        else:
            vmin = field_default_vmin
    if vmax is None:
        if 'valid_max' in field_dict:
            vmax = field_dict['valid_max']
        else:
            vmax = field_default_vmax
    return vmin, vmax


def parse_lon_lat(grid, lon, lat):
    """ Parse lat and lon parameters """
    if lat is None:
        lat = grid.origin_latitude['data'][0]
    if lon is None:
        lon = grid.origin_longitude['data'][0]
    return lon, lat


def generate_colorbar_label(standard_name, units):
    """ Generate and return a label for a colorbar. """
    return str(standard_name).replace('_', ' ') + ' (' + units + ')'


def generate_field_name(container, field):
    """ Return a nice field name for a particular field. """
    if 'standard_name' in container.fields[field]:
        field_name = container.fields[field]['standard_name']
    elif 'long_name' in container.fields[field]:
        field_name = container.fields[field]['long_name']
    else:
        field_name = str(field)
    field_name = field_name.replace('_', ' ')
    field_name = field_name[0].upper() + field_name[1:]
    return field_name


def generate_radar_name(radar):
    """ Return radar name. """
    if 'instrument_name' in radar.metadata:
        return radar.metadata['instrument_name']
    else:
        return ''


def generate_grid_name(grid):
    """ Return grid name. """
    if 'instrument_name' in grid.metadata:
        return grid.metadata['instrument_name']
    else:
        return ''


def generate_radar_time_begin(radar):
    """ Return time begin in datetime instance. """
    # datetime object describing first sweep time
    times = radar.time['data'][0]
    units = radar.time['units']
    calendar = radar.time['calendar']
    return num2date(times, units, calendar)


def generate_radar_time_sweep(radar, sweep):
    """ Return time that a specific sweep began in a datetime instance. """
    first_ray = radar.sweep_start_ray_index['data'][sweep]
    times = radar.time['data'][first_ray]
    units = radar.time['units']
    calendar = radar.time['calendar']
    return num2date(times, units, calendar)


def generate_grid_time_begin(grid):
    """ Return time begin in datetime instance. """
    times = grid.time['data'][0]
    units = grid.time['units']
    if 'calendar' in grid.time:
        calendar = grid.time['calendar']
    else:
        calendar = 'standard'
    return num2date(times, units, calendar)


def generate_filename(radar, field, sweep, ext='png', datetime_format='%Y%m%d%H%M%S', use_sweep_time=False):
    """
    Generate a filename for a plot.

    Generated filename has form:
        radar_name_field_sweep_time.ext

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    sweep : int
        Sweep plotted.
    ext : str
        Filename extension.
    datetime_format : str
        Format of datetime (using strftime format).
    use_sweep_time : bool
        If true, the current sweep's beginning time is used.

    Returns
    -------
    filename : str
        Filename suitable for saving a plot.

    """
    name_s = generate_radar_name(radar).replace(' ', '_')
    field_s = field.replace(' ', '_')
    if use_sweep_time:
        time_s = generate_radar_time_sweep(radar, sweep).strftime(datetime_format)
    else:
        time_s = generate_radar_time_begin(radar).strftime(datetime_format)
    sweep_s = str(sweep).zfill(2)
    return '%s_%s_%s_%s.%s' % (name_s, field_s, sweep_s, time_s, ext)


def generate_grid_filename(grid, field, level, ext='png'):
    """
    Generate a filename for a plot.

    Generated filename has form:
        grid_name_field_level_time.ext

    Parameters
    ----------
    grid : Grid
        Grid structure.
    field : str
        Field plotted.
    level : int
        Level plotted.
    ext : str
        Filename extension.

    Returns
    -------
    filename : str
        Filename suitable for saving a plot.

    """
    name_s = generate_grid_name(grid).replace(' ', '_')
    field_s = field.replace(' ', '_')
    time_s = generate_grid_time_begin(grid).strftime('%Y%m%d%H%M%S')
    level_s = str(level).zfill(2)
    return '%s_%s_%s_%s.%s' % (name_s, field_s, level_s, time_s, ext)


def generate_title(radar, field, sweep, datetime_format=None, use_sweep_time=True):
    """
    Generate a title for a plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    sweep : int
        Sweep plotted.
    datetime_format : str
        Format of datetime (using strftime format).
    use_sweep_time : bool
        If true, the current sweep's beginning time is used.

    Returns
    -------
    title : str
        Plot title.

    """
    if use_sweep_time:
        begin_time = generate_radar_time_sweep(radar, sweep)
    else:
        begin_time = generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + 'Z'
    fixed_angle = radar.fixed_angle['data'][sweep]
    l1 = "%s %.1f Deg. %s " % (generate_radar_name(radar), fixed_angle,
                               time_str)
    field_name = generate_field_name(radar, field)
    return l1 + '\n' + field_name


def generate_grid_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Verical level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).isoformat() + 'Z'
    height = grid.z['data'][level] / 1000.
    l1 = "%s %.1f km %s " % (generate_grid_name(grid), height,
                             time_str)
    field_name = generate_field_name(grid, field)
    return l1 + '\n' + field_name


def generate_longitudinal_level_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Longitudinal level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).isoformat() + 'Z'
    disp = grid.x['data'][level] / 1000.
    if disp >= 0:
        direction = "east"
    else:
        direction = "west"
        disp = - disp
    l1 = "%s %.1f km %s of origin %s " % (generate_grid_name(grid), disp,
                                          direction, time_str)
    field_name = generate_field_name(grid, field)
    return l1 + '\n' + field_name


def generate_latitudinal_level_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Latitudinal level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).isoformat() + 'Z'
    disp = grid.y['data'][level] / 1000.
    if disp >= 0:
        direction = "north"
    else:
        direction = "south"
        disp = - disp
    l1 = "%s %.1f km %s of origin %s " % (generate_grid_name(grid), disp,
                                          direction, time_str)
    field_name = generate_field_name(grid, field)
    return l1 + '\n' + field_name


def generate_vpt_title(radar, field):
    """
    Generate a title for a VPT plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + 'Z'
    l1 = "%s %s " % (generate_radar_name(radar), time_str)
    field_name = generate_field_name(radar, field)
    return l1 + '\n' + field_name


def generate_ray_title(radar, field, ray):
    """
    Generate a title for a ray plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    ray : int
        Ray plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + 'Z'
    l1 = "%s %s" % (generate_radar_name(radar), time_str)
    azim = radar.azimuth['data'][ray]
    elev = radar.elevation['data'][ray]
    l2 = "Ray: %i  Elevation: %.1f Azimuth: %.1f" % (ray, elev, azim)
    field_name = generate_field_name(radar, field)
    return l1 + '\n' + l2 + '\n' + field_name


def generate_az_rhi_title(radar, field, azimuth):
    """
    Generate a title for a pseudo-RHI from PPI azimuth plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    azimuth : float
        Azimuth plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + 'Z'
    l1 = "%s %s " % (generate_radar_name(radar), time_str)
    l2 = "Azimuth: %.1f deg" % azimuth
    field_name = generate_field_name(radar, field)
    return l1 + '\n' + l2 + '\n' + field_name


def set_limits(xlim=None, ylim=None, ax=None):
    """
    Set the display limits.

    Parameters
    ----------
    xlim : tuple, optional
        2-Tuple containing y-axis limits in km. None uses default limits.
    ylim : tuple, optional
        2-Tuple containing x-axis limits in km. None uses default limits.
    ax : Axis
        Axis to adjust.  None will adjust the current axis.

    """
    if ax is None:
        ax = plt.gca()
    if ylim is not None:
        ax.set_ylim(ylim)
    if xlim is not None:
        ax.set_xlim(xlim)



""" Unit Tests for Py-ART's graph/common.py module. """

import datetime

from numpy.testing import assert_almost_equal

import matplotlib
import matplotlib.pyplot as plt

import pyart
from pyart.graph import common


def test_parse_ax():
    ax = common.parse_ax(None)
    assert isinstance(ax, matplotlib.axes.Axes)

    ax1 = plt.gca()
    ax2 = common.parse_ax(ax1)
    assert ax1 == ax2


def test_parse_ax_fig():
    ax, fig = common.parse_ax_fig(None, None)
    assert isinstance(ax, matplotlib.axes.Axes)
    assert isinstance(fig, matplotlib.figure.Figure)

    ax1 = plt.gca()
    fig1 = plt.gcf()
    ax2, fig2 = common.parse_ax_fig(ax1, fig1)
    assert ax1 == ax2
    assert fig1 == fig2


def test_parse_cmap():
    assert common.parse_cmap('jet', 'foo') == 'jet'
    assert common.parse_cmap(None, 'reflectivity') == 'pyart_NWSRef'


def test_parse_vmin_vmax():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.fields['foo'] = {}

    vmin, vmax = common.parse_vmin_vmax(radar, 'foo', None, None)
    assert vmin is None
    assert vmax is None

    vmin, vmax = common.parse_vmin_vmax(radar, 'foo', 10, 20)
    assert vmin == 10
    assert vmax == 20

    radar.fields['foo']['valid_min'] = 30
    radar.fields['foo']['valid_max'] = 40
    vmin, vmax = common.parse_vmin_vmax(radar, 'foo', None, None)
    assert vmin == 30
    assert vmax == 40


def test_parse_lon_lat():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)

    lon, lat = common.parse_lon_lat(grid, None, None)
    assert_almost_equal(lon, -98.1, 2)
    assert_almost_equal(lat, 36.74, 2)

    lon, lat = common.parse_lon_lat(grid, -12.34, 56.78)
    assert_almost_equal(lon, -12.34, 2)
    assert_almost_equal(lat, 56.78, 2)


def test_generate_colorbar_label():
    label = common.generate_colorbar_label('special_name', 'km')
    assert label == 'special name (km)'


def test_generate_field_name():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.fields['foo'] = {}

    field_name = common.generate_field_name(radar, 'foo')
    assert field_name == 'Foo'

    radar.fields['foo']['long_name'] = 'special_name'
    field_name = common.generate_field_name(radar, 'foo')
    assert field_name == 'Special name'

    radar.fields['foo']['standard_name'] = 'other_name'
    field_name = common.generate_field_name(radar, 'foo')
    assert field_name == 'Other name'


def test_generate_radar_name():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)

    radar.metadata['instrument_name'] = 'foobar'
    name = common.generate_radar_name(radar)
    assert name == 'foobar'

    del radar.metadata['instrument_name']
    name = common.generate_radar_name(radar)
    assert name == ''


def test_generate_grid_name():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)

    grid.metadata['instrument_name'] = 'foobar'
    name = common.generate_grid_name(grid)
    assert name == 'foobar'

    del grid.metadata['instrument_name']
    name = common.generate_grid_name(grid)
    assert name == ''


def test_generate_radar_time_begin():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    time = common.generate_radar_time_begin(radar)
    assert time == datetime.datetime(1989, 1, 1, 0, 0, 1)


def test_generate_radar_time_sweep():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 2)

    time = common.generate_radar_time_sweep(radar, 0)
    assert time == datetime.datetime(1989, 1, 1, 0, 0, 1)

    time = common.generate_radar_time_sweep(radar, 1)
    assert time == datetime.datetime(1989, 1, 1, 0, 0, 2)


def test_generate_grid_time_begin():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)

    time = common.generate_grid_time_begin(grid)
    assert time == datetime.datetime(2000, 1, 1, 0, 0, 0)

    del grid.time['calendar']
    time = common.generate_grid_time_begin(grid)
    assert time == datetime.datetime(2000, 1, 1, 0, 0, 0)


def test_generate_filename():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 2)
    filename = common.generate_filename(radar, 'foobar', 0)
    assert filename == 'fake_radar_foobar_00_19890101000001.png'

    filename = common.generate_filename(radar, 'foobar', 1)
    assert filename == 'fake_radar_foobar_01_19890101000001.png'

    filename = common.generate_filename(radar, 'foobar', 1, datetime_format='%Y%m%d')
    assert filename == 'fake_radar_foobar_01_19890101.png'

    filename = common.generate_filename(radar, 'foobar', 1, use_sweep_time=True)
    assert filename == 'fake_radar_foobar_01_19890101000002.png'


def test_generate_grid_filename():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    grid.metadata['instrument_name'] = 'foo'

    filename = common.generate_grid_filename(grid, 'bar', 0)
    assert filename == 'foo_bar_00_20000101000000.png'


def test_generate_title():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 2)
    radar.fields['foo'] = {}

    title = common.generate_title(radar, 'foo', 0)
    assert title == 'fake_radar 0.8 Deg. 1989-01-01T00:00:01Z \nFoo'

    title = common.generate_title(radar, 'foo', 1)
    assert title == 'fake_radar 0.8 Deg. 1989-01-01T00:00:02Z \nFoo'

    title = common.generate_title(radar, 'foo', 1, datetime_format='%Y-%m-%d')
    assert title == 'fake_radar 0.8 Deg. 1989-01-01 \nFoo'

    title = common.generate_title(radar, 'foo', 1, use_sweep_time=False)
    assert title == 'fake_radar 0.8 Deg. 1989-01-01T00:00:01Z \nFoo'


def test_generate_grid_title():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    grid.metadata['instrument_name'] = 'bar'
    grid.fields['foo'] = {}

    title = common.generate_grid_title(grid, 'foo', 0)
    assert title == 'bar 0.0 km 2000-01-01T00:00:00Z \nFoo'


def test_generate_longitudinal_level_title():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    grid.metadata['instrument_name'] = 'bar'
    grid.fields['foo'] = {}

    title = common.generate_longitudinal_level_title(grid, 'foo', 0)
    assert title == (
        'bar 0.0 km east of origin 2000-01-01T00:00:00Z \nFoo')

    grid.x['data'][0] = -1000.
    title = common.generate_longitudinal_level_title(grid, 'foo', 0)
    assert title == (
        'bar 1.0 km west of origin 2000-01-01T00:00:00Z \nFoo')


def test_generate_latitudinal_level_title():
    grid_shape = (1, 1, 1)
    grid_limits = ((0, 1), (0, 1), (0, 1))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    grid.metadata['instrument_name'] = 'bar'
    grid.fields['foo'] = {}

    title = common.generate_latitudinal_level_title(grid, 'foo', 0)
    assert title == (
        'bar 0.0 km north of origin 2000-01-01T00:00:00Z \nFoo')

    grid.y['data'][0] = -1000.
    title = common.generate_latitudinal_level_title(grid, 'foo', 0)
    assert title == (
        'bar 1.0 km south of origin 2000-01-01T00:00:00Z \nFoo')


def test_generate_vpt_title():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.fields['foo'] = {}

    title = common.generate_vpt_title(radar, 'foo')
    assert title == 'fake_radar 1989-01-01T00:00:01Z \nFoo'


def test_generate_ray_title():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.fields['foo'] = {}

    title = common.generate_ray_title(radar, 'foo', 0)
    assert title == ('fake_radar 1989-01-01T00:00:01Z\n' +
                     'Ray: 0  Elevation: 0.8 Azimuth: 0.0\nFoo')


def test_generate_az_rhi_title():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.fields['foo'] = {}

    title = common.generate_az_rhi_title(radar, 'foo', 0)
    assert title == (
        'fake_radar 1989-01-01T00:00:01Z \nAzimuth: 0.0 deg\nFoo')


def test_set_limits():
    common.set_limits(xlim=(-20, 20), ylim=(-10, 10))
    ax = plt.gca()
    assert ax.get_ylim() == (-10, 10)
    assert ax.get_xlim() == (-20, 20)



"""
====================================================
Create a plot of multiple moments from a NEXRAD file
====================================================

An example which creates a plot containing multiple moments taken from a
NEXRAD Archive file.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart

filename = 'KATX20130717_195021_V06'
radar = pyart.io.read_nexrad_archive(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=(10, 10))

ax = fig.add_subplot(221)
display.plot('velocity', 1, ax=ax, title='Doppler Velocity',
             colorbar_label='',
             axislabels=('', 'North South distance from radar (km)'))
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(222)
display.plot('differential_reflectivity', 0, ax=ax,
             title='Differential Reflectivity', colorbar_label='',
             axislabels=('', ''))
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(223)
display.plot('differential_phase', 0, ax=ax,
             title='Differential Phase', colorbar_label='')
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(224)
display.plot('cross_correlation_ratio', 0, ax=ax,
             title='Correlation Coefficient', colorbar_label='',
             axislabels=('East West distance from radar (km)', ''))
display.set_limits((-300, 300), (-300, 300), ax=ax)

plt.show()



"""
==============================
Create a three panel grid plot
==============================

An example which creates a three-panel plot of a gridded NEXRAD radar on a map
with latitude and longitude slices of the reflectivity.  The NCEP
North American regional reanalysis (NARR) pressure is plotted on top of the
grid.

"""
print(__doc__)

# Author Jonathan J. Helmus
# License: BSD 3 clause

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from netCDF4 import num2date, date2num, Dataset
import pyart


# read in the NEXRAD data, create the display
fname = '20110520100000_nexrad_grid.nc'
grid = pyart.io.read_grid(fname)
display = pyart.graph.GridMapDisplay(grid)

# create the figure
font = {'size': 16}
matplotlib.rc('font', **font)
fig = plt.figure(figsize=[15, 8])

# panel sizes
map_panel_axes = [0.05, 0.05, .4, .80]
x_cut_panel_axes = [0.55, 0.10, .4, .30]
y_cut_panel_axes = [0.55, 0.50, .4, .30]
colorbar_panel_axes = [0.05, 0.90, .4, .03]

# parameters
level = 5
vmin = -8
vmax = 64
lat = 36.5
lon = -98.5

# panel 1, basemap, radar reflectivity and NARR overlay
ax1 = fig.add_axes(map_panel_axes)
display.plot_basemap()
display.plot_grid('REF', level=level, vmin=vmin, vmax=vmax, title_flag=False,
                  colorbar_flag=False)
display.plot_crosshairs(lon=lon, lat=lat)

# fetch NCEP NARR data
grid_date = num2date(grid.time['data'], grid.time['units'])[0]
y_m_d = grid_date.strftime('%Y%m%d')
y_m = grid_date.strftime('%Y%m')
url = ('http://nomads.ncdc.noaa.gov/dods/NCEP_NARR_DAILY/' + y_m + '/' +
       y_m_d + '/narr-a_221_' + y_m_d + '_0000_000')
# Use a local copy of the online NCEP NARR data, this file can be created with
# the command:
# nccopy http://nomads.ncdc.noaa.gov/dods/NCEP_NARR_DAILY/201105/20110520/narr-a_221_20110520_0000_000?lon,lat,time,prmsl narr-a_221_20110520_0000_000.nc
# comment out the next line to retrieve the data from the OPeNDAP server.
url = 'narr-a_221_20110520_0000_000.nc'
data = Dataset(url)

# extract data at correct time
data_time = data.variables['time']
t_idx = abs(data_time[:] - date2num(grid_date, data_time.units)).argmin()
prmsl = 0.01 * data.variables['prmsl'][t_idx]

# plot the reanalysis on the basemap
lons, lats = np.meshgrid(data.variables['lon'], data.variables['lat'][:])
x, y = display.basemap(lons, lats)
clevs = np.arange(900, 1100., 1.)
display.basemap.contour(x, y, prmsl, clevs, colors='k', linewidths=1.)

# colorbar
cbax = fig.add_axes(colorbar_panel_axes)
display.plot_colorbar(cax=cbax)

# panel 2, longitude slice.
ax2 = fig.add_axes(x_cut_panel_axes)
display.plot_longitude_slice(
    'REF', lon=lon, lat=lat, vmin=vmin, vmax=vmax, title_flag=False,
    colorbar_flag=False, edges=False,
    axislabels=('Distance from SGP CF (km)', 'Height (km)'))

# panel 3, latitude slice
ax3 = fig.add_axes(y_cut_panel_axes)
display.plot_latitude_slice(
    'REF', lon=lon, lat=lat, vmin=vmin, vmax=vmax, title_flag=False,
    colorbar_flag=False, edges=False,
    axislabels=('', 'Height (km)'))

# add a title
slc_height = grid.z['data'][level]
dts = num2date(grid.time['data'], grid.time['units'])
datestr = dts[0].strftime('%H:%M Z on %Y-%m-%d')
title = 'Sliced at ' + str(slc_height) + ' meters at ' + datestr
fig.text(0.5, 0.9, title)
plt.show()
